import re
from os import path, system
from pathlib import Path
from subprocess import check_output

from yaralyzer.util.logging import log

from pdfalyzer.config import PDF_PARSER_EXECUTABLE_ENV_VAR, PdfalyzerConfig

# PDF Internal Data Regexes
PDF_OBJECT_START_REGEX = re.compile('^obj (\\d+) \\d+$')
CONTAINS_STREAM_REGEX = re.compile('\\s+Contains stream$')

# Install info
DIDIER_STEVENS_RAW_GITHUB_URL = 'https://raw.githubusercontent.com/DidierStevens/DidierStevensSuite/master/'
PDF_PARSER_GITHUB_URL = DIDIER_STEVENS_RAW_GITHUB_URL + 'pdf-parser.py'
PDF_PARSER_INSTALL_MSG = f"If you need to install pdf-parser.py, it's a single .py file that can be " + \
                          "found at '{PDF_PARSER_GITHUB_URL}'."


class PdfParserManager:
    """Instances of this class manage external calls to Didier Stevens's pdf-parser.py for a given PDF."""
    def __init__(self, path_to_pdf: str | Path):
        if PdfalyzerConfig.PDF_PARSER_EXECUTABLE is None:
            raise RuntimeError(f"{PDF_PARSER_EXECUTABLE_ENV_VAR} not configured.\n\n{PDF_PARSER_INSTALL_MSG}")

        if not path.exists(PdfalyzerConfig.PDF_PARSER_EXECUTABLE):
            msg = f"pdf-parser.py not found at configured location '{PdfalyzerConfig.PDF_PARSER_EXECUTABLE}'\n\n"
            msg += PDF_PARSER_INSTALL_MSG
            raise RuntimeError(msg)

        self.path_to_pdf = path_to_pdf
        self.base_shell_cmd = f'{PdfalyzerConfig.PDF_PARSER_EXECUTABLE} -O "{path_to_pdf}"'
        self.object_ids = []
        self.object_ids_containing_stream_data = []
        self.extract_object_ids()

    def extract_object_ids(self) -> None:
        """Examine output of pdf-parser.py to find all object IDs as well as those object IDs that have streams"""
        log.debug(f"Running '{self.base_shell_cmd}'")
        self.pdf_parser_output_lines = check_output(self.base_shell_cmd, shell=True, text=True).split("\n")
        current_object_id = None

        for line in self.pdf_parser_output_lines:
            match = PDF_OBJECT_START_REGEX.match(line)

            if match:
                current_object_id = int(match[1])
                self.object_ids.append(current_object_id)

            if current_object_id is None:
                continue

            if CONTAINS_STREAM_REGEX.match(line):
                log.debug(f"{current_object_id} contains a stream!")
                self.object_ids_containing_stream_data.append(current_object_id)

        log.info(f"{self.path_to_pdf} Object IDs: {self.object_ids}")
        log.info(f"{self.path_to_pdf} Objs IDs w/streams: {self.object_ids_containing_stream_data}")

    def extract_all_streams(self, output_dir: str | Path) -> None:
        """Use pdf-parser.py to find binary data streams in the PDF and dump each of them to a separate file"""
        for object_id in self.object_ids_containing_stream_data:
            stream_dump_file = path.join(output_dir, f'{path.basename(self.path_to_pdf)}.object_{object_id}.bin')
            shell_cmd = self.base_shell_cmd + f' -f -o {object_id} -d "{stream_dump_file}"'
            log.debug(f'Dumping stream from object {object_id}: {shell_cmd}')
            system(shell_cmd)



"""
Decorator for PyPDF PdfObject that extracts a couple of properties (type, label, etc).
"""
from dataclasses import dataclass
from typing import Any, List, Optional, Union

from pypdf.generic import DictionaryObject, IndirectObject, NumberObject, PdfObject
from rich.text import Text
from yaralyzer.util.logging import log

from pdfalyzer.helpers.pdf_object_helper import pypdf_class_name
from pdfalyzer.helpers.rich_text_helper import comma_join_txt, node_label
from pdfalyzer.helpers.string_helper import root_address
from pdfalyzer.output.styles.node_colors import get_class_style, get_class_style_dim
from pdfalyzer.util.adobe_strings import *


@dataclass
class PdfObjectProperties:
    """Simple class to extract critical features of a `PdfObject`."""
    obj: PdfObject
    address: str
    idnum: int
    indirect_object: IndirectObject | None = None

    def __post_init__(self,):
        self.sub_type = None
        self.all_references_processed = False
        self.known_to_parent_as: Optional[str] = None

        if isinstance(self.obj, DictionaryObject):
            self.type = self.obj.get(TYPE) or self.address
            self.sub_type = self.obj.get(SUBTYPE) or self.obj.get(S)

            if TYPE in self.obj and self.sub_type is not None:
                self.label = f"{self.type}:{self.sub_type[1:]}"
            else:
                self.label = self.type

            if isinstance(self.type, str):
                self.type = root_address(self.type)
                self.label = root_address(self.label)
        else:
            # If it's not a DictionaryObject all we have as far as naming is the address passed in.
            self.label = self.address
            self.type = root_address(self.address) if isinstance(self.address, str) else None

        # Force self.label to be a string. TODO this sucks.
        if isinstance(self.label, int):
            self.label = f"{UNLABELED}[{self.label}]"

        # TODO: this is hacky/temporarily incorrect bc we often don't know the parent when node is being constructed
        if isinstance(self.address, int):
            self.first_address = f"[{self.address}]"
        else:
            self.first_address = self.address

        log.debug(f"Node ID: {self.idnum}, type: {self.type}, subtype: {self.sub_type}, " +
                  f"label: {self.label}, first_address: {self.first_address}")

    @classmethod
    def from_reference(cls, reference: IndirectObject, address: str) -> 'PdfObjectProperties':
        return cls(reference.get_object(), address, reference.idnum, reference)

    @classmethod
    def resolve_references(cls, reference_key: str, obj: PdfObject) -> Any:
        """Recursively build the same data structure except IndirectObjects are resolved to nodes."""
        if isinstance(obj, NumberObject):
            return obj.as_numeric()
        elif isinstance(obj, IndirectObject):
            return cls.from_reference(obj, reference_key)
        elif isinstance(obj, list):
            return [cls.resolve_references(reference_key, item) for item in obj]
        elif isinstance(obj, dict):
            return {k: cls.resolve_references(k, v) for k, v in obj.items()}
        else:
            return obj

    @classmethod
    def to_table_row(
        cls,
        reference_key: str,
        obj: PdfObject,
        is_single_row_table: bool = False
    ) -> List[Union[Text, str]]:
        """PDF object property at `reference_key` becomes a formatted 3-tuple for use in Rich tables."""
        with_resolved_refs = cls.resolve_references(reference_key, obj)

        return [
            Text(f"{reference_key}", style='grey' if isinstance(reference_key, int) else ''),
            # Prefix the Text() obj with an empty string to set unstyled chars to the class style of the object
            # they are in.
            Text('', style=get_class_style(obj)).append_text(cls._obj_to_rich_text(with_resolved_refs)),
            # 3rd col (AKA type(value)) is redundant if it's a TextString/Number/etc. node so we make it empty
            '' if is_single_row_table else Text(pypdf_class_name(obj), style=get_class_style_dim(obj))
        ]

    # TODO: this doesn't recurse...
    @classmethod
    def _obj_to_rich_text(cls, obj: Any) -> Text:
        """Recurse through `obj` and build a `Text` object."""
        if isinstance(obj, dict):
            key_value_pairs = [Text(f"{k}: ").append_text(cls._obj_to_rich_text(v)) for k, v in obj.items()]
            return Text('{').append_text(comma_join_txt(key_value_pairs)).append('}')
        elif isinstance(obj, list):
            items = [cls._obj_to_rich_text(item) for item in obj]
            return Text('[').append_text(comma_join_txt(items)).append(']')
        else:
            return cls._to_text(obj)

    @classmethod
    def _to_text(cls, obj: Any) -> Text:
        """Handles styling non-recursable objects."""
        if isinstance(obj, cls):
            return cls.__rich_without_underline__(obj)
        elif isinstance(obj, str):
            return Text(obj)
        else:
            return Text(str(obj), style=get_class_style(obj))

    def __rich_without_underline__(self) -> Text:
        return node_label(self.idnum, self.label, self.obj, underline=False)

    def __rich__(self) -> Text:
        return node_label(self.idnum, self.label, self.obj)

    def __str__(self) -> str:
        return self.__rich__().plain



"""
Some simple math helpers.
"""


def is_divisible_by(n: int, divisor: int) -> bool:
    """Returns True if 'n' is evenly divisible by 'divisor'."""
    return divmod(n, divisor)[1] == 0


def is_even(n: int) -> bool:
    """Returns True if 'n' is even."""
    return is_divisible_by(n, 2)



import io
from logging import Logger
from os import path
from pathlib import Path
from typing import List, Optional, Union

from pypdf import PdfReader, PdfWriter
from pypdf.errors import DependencyError, EmptyFileError, PdfStreamError
from rich.console import Console
from rich.markup import escape
from rich.panel import Panel
from rich.text import Text
from yaralyzer.output.rich_console import console
from yaralyzer.util.logging import log as yaralyzer_log

from pdfalyzer.helpers.filesystem_helper import create_dir_if_it_does_not_exist, insert_suffix_before_extension
from pdfalyzer.helpers.image_helper import ocr_text
from pdfalyzer.helpers.rich_text_helper import (attention_getting_panel, error_text, mild_warning,
     print_error, stderr_console)
from pdfalyzer.helpers.string_helper import exception_str
from pdfalyzer.util.page_range import PageRange

DEPENDENCY_ERROR_MSG = "Pdfalyzer is missing an optional dependency required to extract text. " + \
                       "Try 'pip install pdfalyzer[extract]'"
DEFAULT_PDF_ERRORS_DIR = Path.cwd().joinpath('pdf_errors')
MIN_PDF_SIZE_TO_LOG_PROGRESS_TO_STDERR = 1024 * 1024 * 20


class PdfFile:
    """
    Wrapper for a PDF file path that provides useful methods and properties.

    Attributes:
        file_path (Path): The path to the PDF file.
        basename (str): The base name of the PDF file (with extension).
        basename_without_ext (str): The base name of the PDF file (without extension).
        dirname (Path): The directory containing the PDF file.
        extname (str): The file extension of the PDF file.
        file_size (int): The size of the file in bytes.
    """

    def __init__(self, file_path: Union[str, Path]) -> None:
        """
        Args:
            file_path (Union[str, Path]): Path to the PDF file.
        """
        self.file_path: Path = Path(file_path)

        if not self.file_path.exists():
            raise FileNotFoundError(f"'{file_path}' is not a valid file or directory.")

        self.dirname = self.file_path.parent
        self.basename: str = path.basename(file_path)
        self.basename_without_ext: str = str(Path(self.basename).with_suffix(''))
        self.extname: str = self.file_path.suffix
        self.file_size = self.file_path.stat().st_size

    def extract_page_range(
        self,
        page_range: PageRange,
        destination_dir: Optional[Path] = None,
        extra_file_suffix: Optional[str] = None
    ) -> Path:
        """
        Extract a range of pages to a new PDF file.

        Args:
            page_range (PageRange): Range of pages to extract.
            destination_dir (Optional[Path]): Directory to save the new PDF file. Defaults to the same
                directory as the source PDF.
            extra_file_suffix (Optional[str]): An optional suffix to append to the new PDF's filename.
                Defaults to the page range suffix.

        Returns:
            Path: The path to the newly created PDF file containing the extracted pages.
        """
        destination_dir = Path(destination_dir or self.dirname)
        create_dir_if_it_does_not_exist(destination_dir)
        pdf_reader = PdfReader(self.file_path)
        page_count = len(pdf_reader.pages)
        file_suffix = page_range.file_suffix()

        if page_count < (page_range.last_page - 1):
            raise ValueError(f"PDF only has {page_count} pages but you asked for pages {page_range}!")

        if extra_file_suffix is not None:
            file_suffix += f"__{extra_file_suffix}"

        extracted_pages_pdf_basename = insert_suffix_before_extension(self.file_path, file_suffix).name
        extracted_pages_pdf_path = destination_dir.joinpath(extracted_pages_pdf_basename)
        console.print(f"Extracting {page_range.file_suffix()} from '{self.file_path}' to '{extracted_pages_pdf_path}'")
        pdf_writer = PdfWriter()

        with open(self.file_path, 'rb') as source_pdf:
            pdf_writer.append(fileobj=source_pdf, pages=page_range.to_tuple())

            with open(extracted_pages_pdf_path, 'wb') as extracted_pages_pdf:
                pdf_writer.write(extracted_pages_pdf)

        console.print(f"Extracted pages to new PDF: '{extracted_pages_pdf_path}'.")
        return extracted_pages_pdf_path

    def extract_text(
        self,
        page_range: Optional[PageRange] = None,
        logger: Optional[Logger] = None,
        print_as_parsed: bool = False
    ) -> Optional[str]:
        """
        Use PyPDF to extract text page by page and use Tesseract to OCR any embedded images.

        Args:
            page_range (Optional[PageRange]): If provided, only extract text from pages in this range.
                Page numbers are 1-indexed. If not provided, extract text from all pages.
            log (Optional[Logger]): If provided, log progress to this logger. Otherwise use default logger.
            print_as_parsed (bool): If True, print each page's text to STDOUT as it is parsed.

        Returns:
            Optional[str]: The extracted text, or None if extraction failed.
        """
        from PIL import Image  # Imported here to avoid hard dependency if not using this method
        log = logger or yaralyzer_log
        log.debug(f"Extracting text from '{self.file_path}'...")
        self._page_numbers_of_errors: List[int] = []
        extracted_pages = []

        try:
            pdf_reader = PdfReader(self.file_path)
            page_count = len(pdf_reader.pages)
            log.debug(f"PDF Page count: {page_count}")

            for page_number, page in enumerate(pdf_reader.pages, start=1):
                if page_range and not page_range.in_range(page_number):
                    self._log_to_stderr(f"Skipping page {page_number}...")
                    continue

                self._log_to_stderr(f"Parsing page {page_number}...")
                page_buffer = Console(file=io.StringIO())
                page_buffer.print(Panel(f"PAGE {page_number}", padding=(0, 15), expand=False))
                page_buffer.print(escape(page.extract_text().strip()))
                image_number = 1

                # Extracting images is a bit fraught (lots of PIL and pypdf exceptions have come from here)
                try:
                    for image_number, image in enumerate(page.images, start=1):
                        image_name = f"Page {page_number}, Image {image_number}"
                        self._log_to_stderr(f"   OCRing {image_name}...", "dim")
                        page_buffer.print(Panel(image_name, expand=False))
                        image_obj = Image.open(io.BytesIO(image.data))
                        image_text = ocr_text(image_obj, f"{self.file_path} ({image_name})")
                        page_buffer.print(escape(image_text or '').strip())
                except (OSError, NotImplementedError, TypeError, ValueError) as e:
                    error_str = exception_str(e)
                    msg = f"{error_str} while parsing embedded image {image_number} on page {page_number}..."
                    mild_warning(msg)

                    # Dump an error PDF and encourage user to report to pypdf team.
                    if 'JBIG2Decode' not in str(e):
                        stderr_console.print_exception()

                        if page_number not in self._page_numbers_of_errors:
                            self._handle_extraction_error(page_number, error_str)
                            self._page_numbers_of_errors.append(page_number)

                page_text = page_buffer.file.getvalue()
                extracted_pages.append(page_text)
                log.debug(page_text)

                if print_as_parsed:
                    print(f"{page_text}")
        except DependencyError:
            log.error(DEPENDENCY_ERROR_MSG)
        except EmptyFileError:
            log.warning("Skipping empty file!")
        except PdfStreamError as e:
            stderr_console.print_exception()
            print_error(f"Error parsing PDF file '{self.file_path}': {e}")

        return "\n\n".join(extracted_pages).strip()

    def print_extracted_text(self, page_range: Optional[PageRange] = None, print_as_parsed: bool = False) -> None:
        """Fancy wrapper for printing the extracted text to the screen."""
        console.print(Panel(str(self.file_path), expand=False, style='bright_white reverse'))
        txt = self.extract_text(page_range=page_range, print_as_parsed=print_as_parsed)

        if not print_as_parsed:
            console.print(txt)

    def _handle_extraction_error(self, page_number: int, error_msg: str) -> None:
        """Rip the offending page to a new file and suggest that user report bug to PyPDF."""
        destination_dir = DEFAULT_PDF_ERRORS_DIR

        try:
            extracted_file = self.extract_page_range(PageRange(str(page_number)), destination_dir, error_msg)
        except Exception:
            stderr_console.print_exception()
            stderr_console.print(error_text(f"Failed to extract a page for submission to PyPDF team."))
            extracted_file = None

        blink_txt = Text('', style='bright_white')
        blink_txt.append("An error (", style='blink color(154)').append(error_msg, style='color(11) blink')
        blink_txt.append(') ', style='blink color(154)')
        blink_txt.append("was encountered while processing a PDF file.\n\n", style='blink color(154)')

        txt = Text(f"The error was of a type such that it probably came from a bug in ", style='bright_white')
        txt.append('PyPDF', style='underline bright_green').append('. It was encountered processing the file ')
        txt.append(str(self.file_path), style='file').append('. You should see a stack trace above this box.\n\n')

        txt.append('The offending page will be extracted to ', style='bright_white')
        txt.append(str(extracted_file), style='file').append('.\n\n')
        txt.append(f"Please visit 'https://github.com/py-pdf/pypdf/issues' to report a bug. ", style='bold')
        txt.append(f"Providing the devs with the extracted page and the stack trace help improve pypdf.")
        stderr_console.print(attention_getting_panel(blink_txt + txt, title='PyPDF Error'))

    def _log_to_stderr(self, msg: str, style: Optional[str] = None) -> None:
        """When parsing very large PDFs it can be useful to log progress and other messages to STDERR."""
        if self.file_size < MIN_PDF_SIZE_TO_LOG_PROGRESS_TO_STDERR:
            return

        stderr_console.print(msg, style=style or "")



from dataclasses import dataclass
from typing import Callable, List, Optional

from rich.markup import escape
from yaralyzer.helpers.string_helper import comma_join
from yaralyzer.util.logging import log

from pdfalyzer.decorators.pdf_tree_node import PdfTreeNode
from pdfalyzer.helpers.string_helper import all_strings_are_same_ignoring_numbers, has_a_common_substring
from pdfalyzer.util.adobe_strings import *


@dataclass
class IndeterminateNode:
    """
    Class to handle choosing among the candidates for a given PDF object's parent node.

    Some nodes cannot be placed until we have walked the rest of the tree. For instance
    if we encounter a /Page that relationships /Resources we need to know if there's a
    /Pages parent of the /Page before committing to a tree structure.
    """
    node: PdfTreeNode

    def place_node(self) -> None:
        """Attempt to find the appropriate parent/child relationships for this node."""
        log.debug(f"Attempting to resolve indeterminate node: {self.node}")

        if self._check_for_common_ancestor():
            return
        elif self._check_single_relation_rules():
            return

        # At this point we will choose the relationship with the most descendants as the parent.
        # We will display a warning if we can't find any reason better than that to choose the parent.
        parent = self.find_node_with_most_descendants()
        parent_str = escape(str(parent))

        # Any if/else branch that doesn't return or raise will decide parent to be the node w/most descendants
        if self._has_only_similar_relationships():
            log.info(f"  Fuzzy match addresses or labels; placing under node w/most descendants: {parent_str}")
        elif self._make_parent_if_one_remains(lambda r: r.from_node.type in PAGE_AND_PAGES):
            log.info("  Found a single /Page or /Pages relationship to make parent")
            return
        elif self.node.type == COLOR_SPACE:
            log.info(f"  Color space node found; placing under node w/most descendants: {parent_str}")
        elif set(self.node.unique_labels_of_referring_nodes()) == set(PAGE_AND_PAGES):
            # Handle an edge case seen in the wild involving a PDF that doesn't conform to the PDF spec
            # in a particular way.
            log.warning(f"  {self.node} seems to be a loose {PAGE}. Linking to first {PAGES}")
            pages_nodes = [n for n in self.node.nodes_with_here_references() if self.node.type == PAGES]
            self.node.set_parent(self.find_node_with_most_descendants(pages_nodes))
            return
        else:
            log.warning(f"  {self.node} parent {parent_str} chosen based on descendant count only")
            self.node.log_non_tree_relationships()

        self.node.set_parent(parent)

    def find_node_with_most_descendants(self, list_of_nodes: Optional[List[PdfTreeNode]] = None) -> PdfTreeNode:
        """Find node with a reference to this one that has the most descendants"""
        list_of_nodes = list_of_nodes or [r.from_node for r in self.node.non_tree_relationships]
        max_descendants = max([node.descendants_count() for node in list_of_nodes])
        return find_node_with_lowest_id([n for n in list_of_nodes if n.descendants_count() == max_descendants])

    def _has_only_similar_relationships(self) -> bool:
        """
        Returns `True` if all the nodes w/references to this one have the same type or if all the
        reference_keys that point to this node are the same.
        """
        unique_refferer_labels = self.node.unique_labels_of_referring_nodes()
        unique_addresses = self.node.unique_addresses()

        # Check addresses and referring node labels to see if they are all the same
        reference_keys_or_nodes_are_same = any([
            all_strings_are_same_ignoring_numbers(_list) or has_a_common_substring(_list)
            for _list in [unique_addresses, unique_refferer_labels]
        ])

        return reference_keys_or_nodes_are_same

    def _check_for_common_ancestor(self) -> bool:
        common_ancestor = self._find_common_ancestor_among_nodes(self.node.nodes_with_here_references())

        if common_ancestor is not None:
            log.info(f"  Found common ancestor: {common_ancestor}")
            self.node.set_parent(common_ancestor)
            return True
        else:
            return False

    # TODO could be static method
    def _find_common_ancestor_among_nodes(self, nodes: List[PdfTreeNode]) -> Optional[PdfTreeNode]:
        """If any of 'nodes' is a common ancestor of the rest of the 'nodes', return it."""
        for possible_ancestor in nodes:
            log.debug(f"  Checking possible common ancestor: {possible_ancestor}")
            other_nodes = [n for n in nodes if n != possible_ancestor]

            # Look for a common ancestor; if there is one choose it as the parent.
            if all(possible_ancestor in node.ancestors for node in other_nodes):
                other_nodes_str = comma_join([str(node) for node in other_nodes])
                log.info(f"{possible_ancestor} is the common ancestor of {other_nodes_str}")
                return possible_ancestor

    def _check_single_relation_rules(self) -> bool:
        """Check various ways of narrowing down the list of potential parents to one node."""
        if self._make_parent_if_one_remains(lambda r: r.reference_key in [K, KIDS]):
            log.info("  Found single explicit /K or /Kids ref")
        elif self._make_parent_if_one_remains(lambda r: r.from_node.type not in NON_TREE_KEYS):
            log.info("  Found single determinate relationship")
        else:
            return False

        return True

    def _make_parent_if_one_remains(self, is_possible_parent: Callable) -> bool:
        """Relationships are filtered w/is_possible_parent(); if there's only one possibility it's made the parent."""
        remaining_relationships = [r for r in self.node.non_tree_relationships if is_possible_parent(r)]

        if len(remaining_relationships) == 1:
            log.info(f"  Single remaining relationship {remaining_relationships[0]}; making it the parent")
            self.node.set_parent(remaining_relationships[0].from_node)
            return True
        else:
            return False


def find_node_with_lowest_id(list_of_nodes: List[PdfTreeNode]) -> PdfTreeNode:
    """Return node in `list_of_nodes` with lowest ID."""
    lowest_idnum = min([n.idnum for n in list_of_nodes])
    return next(n for n in list_of_nodes if n.idnum == lowest_idnum)



"""
PDFalyzer: Analyze and explore the structure of PDF files.
"""
from os.path import basename
from typing import Dict, Iterator, List, Optional

from anytree import LevelOrderIter, SymlinkNode
from anytree.search import findall, findall_by_attr
from pypdf import PdfReader
from pypdf.errors import PdfReadError
from pypdf.generic import IndirectObject
from yaralyzer.helpers.file_helper import load_binary_data
from yaralyzer.helpers.rich_text_helper import print_fatal_error_and_exit
from yaralyzer.output.file_hashes_table import compute_file_hashes
from yaralyzer.output.rich_console import console
from yaralyzer.util.logging import log

from pdfalyzer.decorators.document_model_printer import print_with_header
from pdfalyzer.decorators.indeterminate_node import IndeterminateNode
from pdfalyzer.decorators.pdf_tree_node import PdfTreeNode
from pdfalyzer.decorators.pdf_tree_verifier import PdfTreeVerifier
from pdfalyzer.font_info import FontInfo
from pdfalyzer.helpers.rich_text_helper import print_error
from pdfalyzer.pdf_object_relationship import PdfObjectRelationship
from pdfalyzer.util.adobe_strings import *
from pdfalyzer.util.exceptions import PdfWalkError

TRAILER_FALLBACK_ID = 10_000_000
PYPDF_ERROR_MSG = "Failed to open file with PyPDF. Consider filing a PyPDF bug report: https://github.com/py-pdf/pypdf/issues"


class Pdfalyzer:
    """
    Walks a PDF's internals and builds the PDF logical structure tree.

    Each of the PDF's internal objects isw rapped in a `PdfTreeNode` object. The tree is managed
    by the `anytree` library. Information about the tree as a whole is stored in this class.
    Once the PDF is parsed this class provides access to info about or from the underlying PDF tree.

    Attributes:
        font_infos (List[FontInfo]): Font summary objects
        font_info_extraction_error (Optional[Exception]): Error encountered extracting FontInfo (if any)
        max_generation (int): Max revision number ("generation") encounted in this PDF.
        nodes_encountered (Dict[int, PdfTreeNode]): Nodes we've traversed already.
        pdf_basename (str): The base name of the PDF file (with extension).
        pdf_bytes (bytes): PDF binary data.
        pdf_bytes_info (BytesInfo): File size, hashes, and other data points about the PDF's raw bytes.
        pdf_filehandle (BufferedReader): File handle that reads the PDF.
        pdf_path (str): The path to the PDF file.
        pdf_size (int): Number of nodes as extracted from the PDF's Trailer node.
        pdf_tree (PdfTreeNode): The top node of the PDF data structure tree.
        verifier (PdfTreeVerifier): PdfTreeVerifier that can validate the PDF has been walked successfully.
    """

    def __init__(self, pdf_path: str):
        """
        Args:
            pdf_path: Path to the PDF file to analyze
        """
        self.pdf_path = pdf_path
        self.pdf_basename = basename(pdf_path)
        self.pdf_bytes = load_binary_data(pdf_path)
        self.pdf_bytes_info = compute_file_hashes(self.pdf_bytes)
        self.pdf_filehandle = open(pdf_path, 'rb')  # Filehandle must be left open for PyPDF to perform seeks

        try:
            self.pdf_reader = PdfReader(self.pdf_filehandle)
        except PdfReadError:
            self._handle_fatal_error(f'PdfReadError: "{pdf_path}" doesn\'t seem to be a valid PDF file.')
        except Exception as e:
            console.print_exception()
            self._handle_fatal_error(f"{PYPDF_ERROR_MSG}\n{e}")

        # Initialize tracking variables
        self.font_infos: List[FontInfo] = []  # Font summary objects
        self.font_info_extraction_error: Optional[Exception] = None
        self.max_generation = 0  # PDF revisions are "generations"; this is the max generation encountered
        self.nodes_encountered: Dict[int, PdfTreeNode] = {}  # Nodes we've seen already
        self._indeterminate_ids = set()  # See INDETERMINATE_REF_KEYS comment

        # Bootstrap the root of the tree with the trailer. PDFs are always read trailer first.
        # Technically the trailer has no PDF Object ID but we set it to the /Size of the PDF.
        trailer = self.pdf_reader.trailer
        self.pdf_size = trailer.get(SIZE)
        trailer_id = self.pdf_size if self.pdf_size is not None else TRAILER_FALLBACK_ID
        self.pdf_tree = PdfTreeNode(trailer, TRAILER, trailer_id)
        self.nodes_encountered[self.pdf_tree.idnum] = self.pdf_tree

        # Build tree by recursively following relationships between nodes
        self.walk_node(self.pdf_tree)

        # After scanning all objects we place nodes whose position was uncertain, extract fonts, and verify
        self._resolve_indeterminate_nodes()
        self._extract_font_infos()
        self.verifier = PdfTreeVerifier(self)
        self.verifier.verify_all_nodes_encountered_are_in_tree()
        self.verifier.verify_unencountered_are_untraversable()

        # Create SymlinkNodes for relationships between PDF objects that are not parent/child relationships.
        # (Do this last because it has the side effect of making a lot more nodes)
        for node in self.node_iterator():
            if not isinstance(node, SymlinkNode):
                node.symlink_non_tree_relationships()

        log.info(f"Walk complete.")

    def walk_node(self, node: PdfTreeNode) -> None:
        """Recursively walk the PDF's tree structure starting at a given node."""
        log.info(f'walk_node() called with {node}. Object dump:\n{print_with_header(node.obj, node.label)}')
        nodes_to_walk_next = [self._add_relationship_to_pdf_tree(r) for r in node.references_to_other_nodes()]
        node.all_references_processed = True

        for next_node in [n for n in nodes_to_walk_next if not (n is None or n.all_references_processed)]:
            if not next_node.all_references_processed:
                self.walk_node(next_node)

    def find_node_by_idnum(self, idnum) -> Optional[PdfTreeNode]:
        """Find node with `idnum` in the tree. Return `None` if that node is not reachable from the root."""
        nodes = [
            node for node in findall_by_attr(self.pdf_tree, name='idnum', value=idnum)
            if not isinstance(node, SymlinkNode)
        ]

        if len(nodes) == 0:
            return None
        elif len(nodes) == 1:
            return nodes[0]
        else:
            raise PdfWalkError(f"Too many nodes had id {idnum}: {nodes}")

    def is_in_tree(self, search_for_node: PdfTreeNode) -> bool:
        """Returns true if `search_for_node` is in the tree already."""
        return any([node == search_for_node for node in self.node_iterator()])

    def node_iterator(self) -> Iterator[PdfTreeNode]:
        """Iterate over nodes, grouping them by distance from the root."""
        return LevelOrderIter(self.pdf_tree)

    def stream_nodes(self) -> List[PdfTreeNode]:
        """List of actual nodes (not SymlinkNodes) containing streams sorted by PDF object ID"""
        stream_filter = lambda node: node.contains_stream() and not isinstance(node, SymlinkNode)  # noqa: E731
        return sorted(findall(self.pdf_tree, stream_filter), key=lambda r: r.idnum)

    def _add_relationship_to_pdf_tree(self, relationship: PdfObjectRelationship) -> Optional[PdfTreeNode]:
        """
        Place the `relationship` node in the tree. Returns an optional node that should be
        placed in the PDF node processing queue.
        """
        log.info(f'Assessing relationship {relationship}...')
        was_seen_before = (relationship.to_obj.idnum in self.nodes_encountered)  # Must come before _build_or_find()
        from_node = relationship.from_node
        to_node = self._build_or_find_node(relationship.to_obj, relationship.address)
        self.max_generation = max([self.max_generation, relationship.to_obj.generation or 0])

        # If one is already a parent/child of the other there's nothing to do
        if to_node == from_node.parent or to_node in from_node.children:
            log.debug(f"  {from_node} and {to_node} are already related")
            return None

        # Many branches return None or don't return.
        # If there's an explicit /Parent or /Kids relationship then we know the correct relationship
        if relationship.is_parent or relationship.is_child:
            log.debug(f"  Explicit parent/child link: {relationship}")

            if relationship.is_parent:
                from_node.set_parent(to_node)
            elif to_node.parent is not None:
                # Some StructElem nodes I have seen use /P or /K despire not being the real parent/child
                if relationship.from_node.type.startswith(STRUCT_ELEM):
                    log.info(f"{relationship} fail: {to_node} parent is already {to_node.parent}")
                else:
                    log.warning(f"{relationship} fail: {to_node} parent is already {to_node.parent}")
            else:
                from_node.add_child(to_node)

            # Remove this to_node from inteterminacy now that it's got a child or parent
            if relationship.to_obj.idnum in self._indeterminate_ids:
                log.info(f"  Found {relationship} => {to_node} was marked indeterminate but now placed")
                self._indeterminate_ids.remove(relationship.to_obj.idnum)

        # If the relationship is indeterminate or we've seen the PDF object before, add it as
        # a non-tree relationship for now. An attempt to place the node will be made at the end.
        elif relationship.is_indeterminate or relationship.is_link or was_seen_before:
            to_node.add_non_tree_relationship(relationship)

            # If we already encountered 'to_node' then skip adding it to the queue of nodes to walk
            if was_seen_before:
                if relationship.to_obj.idnum not in self._indeterminate_ids and to_node.parent is None:
                    raise PdfWalkError(f"{relationship} - ref has no parent and is not indeterminate")
                else:
                    log.debug(f"  Already saw {relationship}; not scanning next")
                    return None
            # Indeterminate relationships need to wait until everything has been scanned to be placed
            elif relationship.is_indeterminate or (relationship.is_link and not self.is_in_tree(to_node)):
                log.info(f'  Indeterminate ref {relationship}')
                self._indeterminate_ids.add(to_node.idnum)
            # Link nodes like /Dest are usually just links between nodes
            elif relationship.is_link:
                log.debug(f"  Link ref {relationship}")

        # If no other conditions are met make from_node the parent of to_node
        else:
            from_node.add_child(to_node)

        return to_node

    def _handle_fatal_error(self, msg: str) -> None:
        self.pdf_filehandle.close()
        print_fatal_error_and_exit(msg)

    def _resolve_indeterminate_nodes(self) -> None:
        """Place all indeterminate nodes in the tree. Called after all nodes have been walked."""
        indeterminate_nodes = [self.nodes_encountered[idnum] for idnum in self._indeterminate_ids]
        indeterminate_nodes_string = "\n   ".join([f"{node}" for node in indeterminate_nodes])
        log.info(f"Resolving {len(indeterminate_nodes)} indeterminate nodes: {indeterminate_nodes_string}")

        for node in indeterminate_nodes:
            if node.parent is not None:
                log.info(f"{node} marked indeterminate but has parent: {node.parent}")
                continue

            IndeterminateNode(node).place_node()

    def _extract_font_infos(self) -> None:
        """Extract information about fonts in the tree and place it in `self.font_infos`."""
        for node in self.node_iterator():
            if not (isinstance(node.obj, dict) and RESOURCES in node.obj):
                continue

            log.debug(f"Extracting fonts from node with '{RESOURCES}' key: {node}...")
            known_font_ids = [fi.idnum for fi in self.font_infos]

            try:
                self.font_infos += [
                    fi for fi in FontInfo.extract_font_infos(node.obj)
                    if fi.idnum not in known_font_ids
                ]
            except Exception as e:
                self.font_info_extraction_error = e
                console.line()
                log.warning(f"Failed to extract font information from node: {node} (error: {e})")
                console.line()

    def _build_or_find_node(self, relationship: IndirectObject, relationship_key: str) -> PdfTreeNode:
        """If node in self.nodes_encountered already then return it, otherwise build a node and store it."""
        if relationship.idnum in self.nodes_encountered:
            return self.nodes_encountered[relationship.idnum]

        log.debug(f"Building node for {relationship}")
        new_node = PdfTreeNode.from_reference(relationship, relationship_key)
        self.nodes_encountered[relationship.idnum] = new_node
        return new_node

    def _print_nodes_encountered(self) -> None:
        """Debug method that displays which nodes have already been walked."""
        for i in sorted(self.nodes_encountered.keys()):
            console.print(f'{i}: {self.nodes_encountered[i]}')



import code
import sys
from argparse import Namespace
from os import environ, getcwd, path

from dotenv import load_dotenv
from pypdf import PdfWriter
from pypdf.errors import PdfReadError

# Should be first local import before load_dotenv() (or at least I think it needs to come first)
from pdfalyzer.config import PdfalyzerConfig

# load_dotenv() should be called as soon as possible (before parsing local classes) but not for pytest
if not environ.get('INVOKED_BY_PYTEST', False):
    for dotenv_file in [path.join(dir, '.pdfalyzer') for dir in [getcwd(), path.expanduser('~')]]:
        if path.exists(dotenv_file):
            load_dotenv(dotenv_path=dotenv_file)
            break

from rich.columns import Columns
from rich.panel import Panel
from rich.text import Text
from yaralyzer.helpers.rich_text_helper import prefix_with_style
from yaralyzer.output.file_export import invoke_rich_export
from yaralyzer.output.rich_console import console
from yaralyzer.util.logging import log_and_print

from pdfalyzer.decorators.pdf_file import PdfFile
from pdfalyzer.helpers.filesystem_helper import file_size_in_mb, set_max_open_files
from pdfalyzer.helpers.rich_text_helper import print_highlighted
from pdfalyzer.output.pdfalyzer_presenter import PdfalyzerPresenter
from pdfalyzer.output.styles.rich_theme import PDFALYZER_THEME_DICT
from pdfalyzer.pdfalyzer import Pdfalyzer
from pdfalyzer.util.argument_parser import ask_to_proceed, output_sections, parse_arguments
from pdfalyzer.util.cli_tools_argument_parser import (MAX_QUALITY, parse_combine_pdfs_args,
     parse_pdf_page_extraction_args, parse_text_extraction_args)
from pdfalyzer.util.pdf_parser_manager import PdfParserManager

# For the table shown by running pdfalyzer_show_color_theme
MAX_THEME_COL_SIZE = 35


def pdfalyze():
    """Main entry point for The Pdfalyzer command line tool."""
    args = parse_arguments()
    pdfalyzer = Pdfalyzer(args.file_to_scan_path)
    presenter = PdfalyzerPresenter(pdfalyzer)
    output_basepath = None

    # Binary stream extraction is a special case
    if args.extract_binary_streams:
        log_and_print(f"Extracting binary streams in '{args.file_to_scan_path}' to files in '{args.output_dir}'...")
        PdfParserManager(args.file_to_scan_path).extract_all_streams(args.output_dir)
        log_and_print(f"Binary stream extraction complete, files written to '{args.output_dir}'.\nExiting.\n")
        sys.exit()

    # The method that gets called is related to the argument name. See 'possible_output_sections' list in
    # argument_parser.py. Analysis exports wrap themselves around the methods that actually generate the analyses.
    for (arg, method) in output_sections(args, presenter):
        if args.output_dir:
            output_basepath = PdfalyzerConfig.get_output_basepath(method)
            print(f'Exporting {arg} data to {output_basepath}...')
            console.record = True

        method()

        if args.export_txt:
            invoke_rich_export(console.save_text, output_basepath)
        if args.export_html:
            invoke_rich_export(console.save_html, output_basepath)
        if args.export_svg:
            invoke_rich_export(console.save_svg, output_basepath)

        # Clear the buffer if we have one
        if args.output_dir:
            del console._record_buffer[:]

    # Drop into interactive shell if requested
    if args.interact:
        code.interact(local=locals())

    pdfalyzer.pdf_filehandle.close()


def pdfalyzer_show_color_theme() -> None:
    """Utility method to show pdfalyzer's color theme. Invocable with 'pdfalyzer_show_color_theme'."""
    console.print(Panel('The Pdfalyzer Color Theme', style='reverse'))

    colors = [
        prefix_with_style(name[:MAX_THEME_COL_SIZE], style=str(style)).append(' ')
        for name, style in PDFALYZER_THEME_DICT.items()
        if name not in ['reset', 'repr_url']
    ]

    console.print(Columns(colors, column_first=True, padding=(0, 3)))


def combine_pdfs():
    """
    Script method to combine multiple PDFs into one. Invocable with 'combine_pdfs PDF1 [PDF2...]'.
    Example: https://github.com/py-pdf/pypdf/blob/main/docs/user/merging-pdfs.md
    """
    args = parse_combine_pdfs_args()
    set_max_open_files(args.number_of_pdfs)
    merger = PdfWriter()

    for pdf in args.pdfs:
        try:
            print_highlighted(f"  -> Merging '{pdf}'...", style='dim')
            merger.append(pdf)
        except PdfReadError as e:
            print_highlighted(f"      -> Failed to merge '{pdf}'! {e}", style='red')
            ask_to_proceed()

    # Iterate through pages and compress, lowering image quality if requested
    # See https://pypdf.readthedocs.io/en/latest/user/file-size.html#reducing-image-quality
    for i, page in enumerate(merger.pages):
        if args.image_quality < MAX_QUALITY:
            for j, img in enumerate(page.images):
                print_highlighted(
                    f"  -> Reducing image #{j + 1} quality on page {i + 1} to {args.image_quality}...",
                    style='dim'
                )

                img.replace(img.image, quality=args.image_quality)

        print_highlighted(f"  -> Compressing page {i + 1}...", style='dim')
        page.compress_content_streams()  # This is CPU intensive!

    print_highlighted(f"\nWriting '{args.output_file}'...", style='cyan')
    merger.compress_identical_objects(remove_identicals=True, remove_orphans=True)
    merger.write(args.output_file)
    merger.close()
    txt = Text('').append(f"  -> Wrote ")
    txt.append(str(file_size_in_mb(args.output_file)), style='cyan').append(" megabytes\n")
    print_highlighted(txt)


def extract_pdf_pages() -> None:
    """Extract a range of pages from a PDF to a new PDF."""
    args = parse_pdf_page_extraction_args()
    PdfFile(args.pdf_file).extract_page_range(args.page_range, destination_dir=args.destination_dir)


def extract_pdf_text() -> None:
    """Extract text from a list of file or from all PDF files in a list of directories."""
    args: Namespace = parse_text_extraction_args()
    console.line()

    for file_path in args.files_to_process:
        PdfFile(file_path).print_extracted_text(args.page_range, args.print_as_parsed)
        console.line(2)



"""
`BinaryScanner` class.
"""
from collections import defaultdict
from typing import Iterator, Optional, Tuple

from rich.panel import Panel
from rich.text import Text
from yaralyzer.bytes_match import BytesMatch
from yaralyzer.decoding.bytes_decoder import BytesDecoder
from yaralyzer.encoding_detection.character_encodings import BOMS
from yaralyzer.helpers.bytes_helper import hex_string, print_bytes
from yaralyzer.helpers.string_helper import escape_yara_pattern
from yaralyzer.output.regex_match_metrics import RegexMatchMetrics
from yaralyzer.output.rich_console import BYTES_NO_DIM, console, console_width
from yaralyzer.yara.yara_rule_builder import HEX, REGEX, safe_label
from yaralyzer.yaralyzer import Yaralyzer
from yaralyzer.util.logging import log

from pdfalyzer.config import PdfalyzerConfig
from pdfalyzer.decorators.pdf_tree_node import PdfTreeNode
from pdfalyzer.detection.constants.binary_regexes import (BACKTICK, DANGEROUS_PDF_KEYS_TO_HUNT_ONLY_IN_FONTS,
     DANGEROUS_STRINGS, FRONTSLASH, GUILLEMET, QUOTE_PATTERNS)
from pdfalyzer.helpers.string_helper import generate_hyphen_line
from pdfalyzer.output.layout import print_headline_panel, print_section_sub_subheader
from pdfalyzer.util.adobe_strings import CONTENTS, CURRENTFILE_EEXEC, FONT_FILE_KEYS


class BinaryScanner:
    """
    Class for handling binary data - scanning through it for various suspicious patterns as well as forcing
    various character encodings upon it to see what comes out.
    """

    def __init__(self, _bytes: bytes, owner: PdfTreeNode, label: Optional[Text] = None):
        """
        Args:
            _bytes (bytes): The binary data to be scanned.
            owner (PdfTreeNode): The `PdfTreeNode` that contains this binary data.
            label (Optional[Text]): A rich `Text` label for the binary data (e.g. the PDF object's address).
        """
        self.bytes = _bytes
        self.label = label
        self.owner = owner
        self.stream_length = len(_bytes)

        if label is None and isinstance(owner, PdfTreeNode):
            self.label = owner.__rich__()

        self.suppression_notice_queue = []
        self.regex_extraction_stats = defaultdict(lambda: RegexMatchMetrics())

    def check_for_dangerous_instructions(self) -> None:
        """Scan for all the strings in `DANGEROUS_INSTRUCTIONS` list and decode bytes around them."""
        subheader = "Scanning Binary For Anything That Could Be Described As 'sus'..."
        print_section_sub_subheader(subheader, style=f"bright_red")

        for instruction in DANGEROUS_STRINGS:
            yaralyzer = self._pattern_yaralyzer(instruction, REGEX)  # TODO maybe change REGEX const string?
            yaralyzer.highlight_style = 'bright_red bold'
            self.process_yara_matches(yaralyzer, instruction, force=True)

        # TODO code smell: This check should probably be in the calling code not here in the instance method
        if self.owner.type in FONT_FILE_KEYS:
            log.info(f"{self.owner} is a /FontFile. Scanning for short but dangerous PDF keys...")

            for instruction in DANGEROUS_PDF_KEYS_TO_HUNT_ONLY_IN_FONTS:
                yaralyzer = self._pattern_yaralyzer(instruction, REGEX)
                yaralyzer.highlight_style = 'bright_red bold'
                self.process_yara_matches(yaralyzer, instruction, force=True)

    def check_for_boms(self) -> None:
        """Check the binary data for BOMs."""
        print_section_sub_subheader("Scanning Binary for any BOMs...", style='BOM')

        for bom_bytes, bom_name in BOMS.items():
            yaralyzer = self._pattern_yaralyzer(hex_string(bom_bytes), HEX, bom_name)
            yaralyzer.highlight_style = 'BOM'
            self.process_yara_matches(yaralyzer, bom_name, force=True)

    def force_decode_quoted_bytes(self) -> None:
        """
        Find all strings matching `QUOTE_PATTERNS` (AKA between quote chars) and decode them with various
        encodings. The `--quote-type` arg will limit this decode to just one kind of quote.
        """
        quote_selections = PdfalyzerConfig._args.extract_quoteds

        if len(quote_selections) == 0:
            headline = "Skipping extract/decode of quoted bytes (--extract-quoted is empty)"
            print_section_sub_subheader(headline, style='grey')

        for quote_type in quote_selections:
            if self.owner and self.owner.type == CONTENTS and quote_type in [FRONTSLASH, GUILLEMET]:
                msg = f"Not attempting {quote_type} decode for {CONTENTS} node type..."
                print_headline_panel(msg, style='dim')
                continue

            print_section_sub_subheader(
                f"Forcing Decode of {quote_type.capitalize()} Quoted Strings",
                style=BYTES_NO_DIM
            )

            quote_pattern = QUOTE_PATTERNS[quote_type]
            yaralyzer = self._quote_yaralyzer(quote_pattern, quote_type)
            self.process_yara_matches(yaralyzer, f"{quote_type}_quoted")

    # -------------------------------------------------------------------------------
    # These extraction iterators will iterate over all matches for a specific pattern.
    # YARA rules are written on the fly and then YARA does the matching.
    # -------------------------------------------------------------------------------
    def extract_guillemet_quoted_bytes(self) -> Iterator[Tuple[BytesMatch, BytesDecoder]]:
        """Iterate on all strings surrounded by Guillemet quotes, e.g. «string»."""
        return self._quote_yaralyzer(QUOTE_PATTERNS[GUILLEMET], GUILLEMET).match_iterator()

    def extract_backtick_quoted_bytes(self) -> Iterator[Tuple[BytesMatch, BytesDecoder]]:
        """Returns an interator over all strings surrounded by backticks."""
        return self._quote_yaralyzer(QUOTE_PATTERNS[BACKTICK], BACKTICK).match_iterator()

    def extract_front_slash_quoted_bytes(self) -> Iterator[Tuple[BytesMatch, BytesDecoder]]:
        """Returns an interator over all strings surrounded by front_slashes (hint: regular expressions)."""
        return self._quote_yaralyzer(QUOTE_PATTERNS[FRONTSLASH], FRONTSLASH).match_iterator()

    def print_stream_preview(self, num_bytes=None, title_suffix=None) -> None:
        """Print a preview showing the beginning and end of the embedded stream data."""
        num_bytes = num_bytes or PdfalyzerConfig._args.preview_stream_length or console_width()
        snipped_byte_count = self.stream_length - (num_bytes * 2)
        console.line()

        if snipped_byte_count < 0:
            title = f"ALL {self.stream_length} BYTES IN STREAM"
        else:
            title = f"FIRST AND LAST {num_bytes} BYTES OF {self.stream_length} BYTE STREAM"

        title += title_suffix if title_suffix is not None else ''
        title = f"BEGIN {title}".upper()
        console.print(generate_hyphen_line(title=title), style='dim')

        if snipped_byte_count < 0:
            print_bytes(self.bytes)
        else:
            print_bytes(self.bytes[:num_bytes])
            console.print(f"\n    <...skip {snipped_byte_count} bytes...>\n", style='dim')
            print_bytes(self.bytes[-num_bytes:])

        console.print(generate_hyphen_line(title="END " + title), style='dim')
        console.line()

    def process_yara_matches(self, yaralyzer: Yaralyzer, pattern: str, force: bool = False) -> None:
        """
        Decide whether to attempt to decode the matched bytes and track stats.

        Args:
            yaralyzer (Yaralyzer): The `Yaralyzer` instance to use for finding matches.
            pattern (str): The pattern being searched for (used for stats tracking).
            force (bool): If `True`, decode all matches even if they are very short or very long.
        """
        for bytes_match, decoder in yaralyzer.match_iterator():
            log.debug(f"Trackings match stats for {pattern}, bytes_match: {bytes_match}, is_decodable: {bytes_match.is_decodable()}")  # noqa: E501

            # Send suppressed decodes to a queue and track the reason for the suppression in the stats
            if not (bytes_match.is_decodable() or force):
                self.suppression_notice_queue.append(bytes_match.suppression_notice())
                continue

            # Print out any queued suppressed notices before printing non suppressed matches
            self._print_suppression_notices()
            console.print(decoder)
            self.regex_extraction_stats[pattern].tally_match(decoder)  # TODO: This call must come after print(decoder)

        self._print_suppression_notices()

        # This check initializes the defaultdict for 'pattern'
        if self.regex_extraction_stats[pattern].match_count == 0:
            pass

    def bytes_after_eexec_statement(self) -> bytes:
        """Get the bytes after the 'eexec' demarcation line (if it appears). See Adobe docs for details."""
        return self.bytes.split(CURRENTFILE_EEXEC)[1] if CURRENTFILE_EEXEC in self.bytes else self.bytes

    def _quote_yaralyzer(self, quote_pattern: str, quote_type: str):
        """Helper method to build a Yaralyzer for a `quote_pattern`."""
        label = f"{quote_type}_Quoted"

        if quote_type == GUILLEMET:
            return self._pattern_yaralyzer(quote_pattern, HEX, label, label)
        else:
            return self._pattern_yaralyzer(quote_pattern, REGEX, label, label)

    def _pattern_yaralyzer(
        self,
        pattern: str,
        pattern_type: str,
        rules_label: Optional[str] = None,
        pattern_label: Optional[str] = None
    ) -> Yaralyzer:
        """Build a `yaralyzer` to scan `self.bytes`."""
        return Yaralyzer.for_patterns(
            patterns=[escape_yara_pattern(pattern)],
            patterns_type=pattern_type,
            scannable=self.bytes,
            scannable_label=self.label.plain,
            rules_label=safe_label(rules_label or pattern),  # TODO: maybe safe_label() belongs in yaralyzer
            pattern_label=safe_label(pattern_label or pattern)
        )

    def _print_suppression_notices(self) -> None:
        """Print the notices in queue in a single display panel and then empty the queue."""
        if len(self.suppression_notice_queue) == 0:
            return

        suppression_notices_txt = Text("\n").join([notice for notice in self.suppression_notice_queue])
        panel = Panel(suppression_notices_txt, style='bytes', expand=False)
        console.print(panel)
        self.suppression_notice_queue = []

    def _eexec_idx(self) -> int:
        """Returns the location of `CURRENTFILES_EEXEC` within the binary stream data (or 0 if it's not there)."""
        return self.bytes.find(CURRENTFILE_EEXEC) if CURRENTFILE_EEXEC in self.bytes else 0



"""
Methods to help with the formatting of the output tables, headers, panels, etc.
"""
from typing import List

from rich import box
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from yaralyzer.output.rich_console import console, console_width

from pdfalyzer.helpers.rich_text_helper import indent_padding

DEFAULT_SUBTABLE_COL_STYLES = ['white', 'bright_white']
HEADER_PADDING = (1, 1)


def generate_subtable(cols: List[str], header_style: str = 'subtable') -> Table:
    """Suited for placement in larger tables."""
    table = Table(
        box=box.SIMPLE,
        show_edge=False,
        collapse_padding=True,
        padding=(0, 0, 0, 0),
        header_style=header_style,
        show_lines=False,
        border_style='grey.dark',
        expand=True)

    for i, col in enumerate(cols):
        if i == 0:
            table.add_column(col, style=DEFAULT_SUBTABLE_COL_STYLES[0], justify='left')
        else:
            table.add_column(col, style='bright_white', justify='right')

    return table


def subheading_width() -> int:
    """Return 75% of the console width."""
    return int(console_width() * 0.75)


def half_width() -> int:
    """Return 50% of the console width."""
    return int(console_width() * 0.5)


def pad_header(header: str) -> Padding:
    """Would pad anything, not just headers"""
    return Padding(header, HEADER_PADDING)


def print_section_header(headline: str, style: str = '') -> None:
    """Prints a full-width section header with padding above and below."""
    console.line(2)
    _print_header_panel(headline, f"{style} reverse", True, console_width(), HEADER_PADDING)
    console.line()


def print_section_subheader(headline: str, style: str = '') -> None:
    """Prints a half-width section subheader with padding above."""
    console.line()
    _print_header_panel(headline, style, True, subheading_width(), HEADER_PADDING)


def print_section_sub_subheader(headline: str, style: str = ''):
    """Prints a half-width section sub-subheader with no padding above."""
    console.line()
    _print_header_panel(headline, style, True, half_width())


def print_headline_panel(headline: str, style: str = '', indent: int = 0, internal_padding: tuple | None = None):
    """Prints a full-width headline panel with no padding above or below."""
    _print_header_panel(headline, style, False, console_width(), internal_padding, indent=indent)


def print_fatal_error_panel(headline: str):
    """Prints a full-width red blinking panel for fatal errors."""
    print_headline_panel(headline, style='red blink')


def _print_header_panel(
        headline: str,
        style: str,
        expand: bool,
        width: int,
        internal_padding: tuple | None = None,
        indent: int = 0
    ) -> None:
    """Helper to print a rich `Panel` with the given style, width, and padding."""
    panel = Panel(
        headline,
        style=style,
        expand=expand,
        width=width or subheading_width(),
        padding=internal_padding or (0,),
    )

    console.print(Padding(panel, indent_padding(indent)))



"""
Some methods to help with the direct manipulation/processing of PyPDF's PdfObjects
"""
from typing import List, Optional

from pypdf.generic import IndirectObject, PdfObject

from pdfalyzer.pdf_object_relationship import PdfObjectRelationship


def does_list_have_any_references(_list) -> bool:
    """Return true if any element of _list is an IndirectObject."""
    return any(isinstance(item, IndirectObject) for item in _list)


def pdf_object_id(pdf_object) -> Optional[int]:
    """Return the ID of an IndirectObject and None for everything else"""
    return pdf_object.idnum if isinstance(pdf_object, IndirectObject) else None


def pypdf_class_name(obj: PdfObject) -> str:
    """Shortened name of type(obj), e.g. PyPDF.generic._data_structures.ArrayObject becomes Array"""
    class_pkgs = type(obj).__name__.split('.')
    class_pkgs.reverse()
    return class_pkgs[0].removesuffix('Object')


def _sort_pdf_object_refs(refs: List[PdfObjectRelationship]) -> List[PdfObjectRelationship]:
    """Sort a list of PdfObjectRelationship objects by their to_obj's idnum. Only used by pytest."""
    return sorted(refs, key=lambda ref: ref.to_obj.idnum)



"""
Functions for miscellaneous Rich text/string pretty printing operations.
"""
from sys import stderr
from typing import List, Optional, Union

from pypdf.generic import PdfObject
from rich.console import Console
from rich.panel import Panel
from rich.padding import Padding
from rich.text import Text
from yaralyzer.output.rich_console import console

from pdfalyzer.helpers.pdf_object_helper import pypdf_class_name
from pdfalyzer.output.styles.node_colors import get_label_style, get_class_style_italic

ARROW_BULLET = '➤ '

# Usually we use the yaralyzer console but that has no highlighter
pdfalyzer_console = Console(color_system='256')
stderr_console = Console(color_system='256', file=stderr)


def attention_getting_panel(text: Text, title: str, style: str = 'white on red') -> Padding:
    p = Panel(text, padding=(2), title=title, style=style)
    return Padding(p, pad=(1, 10, 2, 10))


def bullet_text(msg: Union[str, Text], style: Optional[str] = None) -> Text:
    if isinstance(msg, str):
        msg = Text(msg, style=style)

    return Text(ARROW_BULLET).append(msg)


def comma_join_txt(text_objs: List[Text]) -> Text:
    return Text(", ").join(text_objs)


def error_text(text: Union[str, Text]) -> Text:
    msg = Text('').append(f"ERROR", style='bright_red').append(": ")

    if isinstance(text, Text):
        return msg + text
    else:
        return msg.append(text)


def indented_bullet(msg: Union[str, Text], style: Optional[str] = None) -> Text:
    return Text('  ') + bullet_text(msg, style)


def indent_padding(indent: int) -> tuple[int, int, int, int]:
    return (0, 0, 0, indent)


def mild_warning(msg: str) -> None:
    console.print(indented_bullet(Text(msg, style='mild_warning')))


def node_label(idnum: int, label: str, pdf_object: PdfObject, underline: bool = True) -> Text:
    """Colored text representation of a PDF node. Example: <5:FontDescriptor(Dictionary)>."""
    text = Text('<', style='white')
    text.append(f'{idnum}', style='bright_white')
    text.append(':', style='white')
    text.append(label[1:], style=f"{get_label_style(label)} {'underline' if underline else ''} bold")
    text.append('(', style='white')
    text.append(pypdf_class_name(pdf_object), style=get_class_style_italic(pdf_object))
    text.append(')', style='white')
    text.append('>')
    return text


def number_and_pct(_number: int, total: int, digits: int = 1) -> Text:
    """Return e.g. '8 (80%)'."""
    return Text(str(_number), style='bright_white').append_text(pct_txt(_number, total, digits))


def pct_txt(_number: int, total: int, digits: int = 1) -> Text:
    """Return nicely formatted percentage, e.g. '(80%)'."""
    pct = (100 * float(_number) / float(total)).__round__(digits)
    return Text(f"({pct}%)", style='blue')


def print_error(text: Union[str, Text]) -> Text:
    console.line()
    console.print(error_text(text))


def print_highlighted(msg: Union[str, Text], **kwargs) -> None:
    """Print 'msg' with Rich highlighting."""
    pdfalyzer_console.print(msg, highlight=True, **kwargs)


def print_warning(text: Union[str, Text]) -> None:
    console.print(_warning_text(text))


def quoted_text(
    _string: str,
    style: str = '',
    quote_char_style: str = 'white',
    quote_char: str = "'"
) -> Text:
    """Wrap _string in 'quote_char'. Style 'quote_char' with 'quote_char_style'."""
    quote_char_txt = Text(quote_char, style=quote_char_style)
    txt = quote_char_txt + Text(_string, style=style) + quote_char_txt
    txt.justify = 'center'
    return txt


def _warning_text(text: Union[str, Text]) -> Text:
    msg = Text('').append(f"WARNING", style='bright_yellow').append(": ")

    if isinstance(text, Text):
        return msg + text
    else:
        return msg.append(text)



"""
Functions to help with the pre-configured YARA rules in the /yara directory.
"""
from importlib.resources import as_file, files
from typing import Optional, Union

from yaralyzer.config import YaralyzerConfig
from yaralyzer.helpers.rich_text_helper import print_fatal_error_and_exit
from yaralyzer.yaralyzer import Yaralyzer

from pdfalyzer.config import PDFALYZER

YARA_RULES_DIR = files(PDFALYZER).joinpath('yara_rules')

YARA_RULES_FILES = [
    'didier_stevens.yara',
    'lprat.static_file_analysis.yara',
    'PDF.yara',
    'PDF_binary_stream.yara',
    'pdf_malware.yara',
]


def get_file_yaralyzer(file_path_to_scan: str) -> Yaralyzer:
    """Get a yaralyzer for a file path."""
    return _build_yaralyzer(file_path_to_scan)


def get_bytes_yaralyzer(scannable: bytes, label: str) -> Yaralyzer:
    """Get a yaralyzer for a `scannable` bytes."""
    return _build_yaralyzer(scannable, label)


def _build_yaralyzer(scannable: Union[bytes, str], label: Optional[str] = None) -> Yaralyzer:
    """Build a yaralyzer for .yara rules files stored in the yara_rules/ dir in this package."""
    # TODO: ugh this sucks (handling to extract .yara files from a python pkg zip)
    with as_file(YARA_RULES_DIR.joinpath(YARA_RULES_FILES[0])) as yara0:
        with as_file(YARA_RULES_DIR.joinpath(YARA_RULES_FILES[1])) as yara1:
            with as_file(YARA_RULES_DIR.joinpath(YARA_RULES_FILES[2])) as yara2:
                with as_file(YARA_RULES_DIR.joinpath(YARA_RULES_FILES[3])) as yara3:
                    with as_file(YARA_RULES_DIR.joinpath(YARA_RULES_FILES[4])) as yara4:
                        # If there is a custom yara_rules arg, use that instead of the files in the yara_rules/ dir
                        rules_paths = YaralyzerConfig.args.yara_rules_files or []

                        if not YaralyzerConfig.args.no_default_yara_rules:
                            rules_paths += [str(y) for y in [yara0, yara1, yara2, yara3, yara4]]

                        try:
                            return Yaralyzer.for_rules_files(rules_paths, scannable, label)
                        except FileNotFoundError as e:
                            print_fatal_error_and_exit(str(e))



"""
Some helpers for stuff with the local filesystem.
"""
import re
from pathlib import Path
from typing import Optional, Union

from yaralyzer.output.rich_console import console
from yaralyzer.util.logging import log

from pdfalyzer.helpers.rich_text_helper import print_highlighted

NUMBERED_PAGE_REGEX = re.compile(r'.*_(\d+)\.\w{3,4}$')
DEFAULT_MAX_OPEN_FILES = 256  # macOS default
OPEN_FILES_BUFFER = 30        # we might have some files open already so we need to go beyond DEFAULT_MAX_OPEN_FILES
PDF_EXT = '.pdf'

# TODO: this kind of type alias is not supported until Python 3.12
# type StrOrPath = Union[str, Path]


def create_dir_if_it_does_not_exist(dir: Path) -> None:
    """Like it says on the tin."""
    if dir.exists():
        return

    log.warning(f"Need to create '{dir}'")
    dir.mkdir(parents=True, exist_ok=True)


def insert_suffix_before_extension(file_path: Path, suffix: str, separator: str = '__') -> Path:
    """Inserting 'page 1' suffix in 'path/to/file.jpg' -> '/path/to/file__page_1.jpg'."""
    suffix = strip_bad_chars(suffix).replace(' ', '_')
    file_path_without_extension = file_path.with_suffix('')
    return Path(f"{file_path_without_extension}{separator}{suffix}{file_path.suffix}")


def is_pdf(file_path: Union[str, Path]) -> bool:
    """Return True if 'file_path' ends with '.pdf'."""
    return str(file_path).endswith(PDF_EXT)


def file_exists(file_path: Union[str, Path]) -> bool:
    """Return True if 'file_path' exists."""
    return Path(file_path).exists()


def do_all_files_exist(file_paths: list[Union[str, Path]]) -> bool:
    """Print an error for each element of 'file_paths' that's not a file. Return True if all 'file_paths' exist."""
    all_files_exist = True

    for file_path in file_paths:
        if not file_exists(file_path):
            console.print(f"File not found: '{file_path}'", style='error')
            all_files_exist = False

    return all_files_exist


def extract_page_number(file_path: Union[str, Path]) -> Optional[int]:
    """Extract the page number from the end of a filename if it exists."""
    match = NUMBERED_PAGE_REGEX.match(str(file_path))
    return int(match.group(1)) if match else None


def file_size_in_mb(file_path: Union[str, Path], decimal_places: int = 2) -> float:
    """Return the size of 'file_path' in MB rounded to 2 decimal places,"""
    return round(Path(file_path).stat().st_size / 1024.0 / 1024.0, decimal_places)


def set_max_open_files(num_filehandles: int = DEFAULT_MAX_OPEN_FILES) -> tuple[Optional[int], Optional[int]]:
    """
    Sets the OS level max open files to at least 'num_filehandles'. Current value can be seen with 'ulimit -a'.
    Required when you might be opening more than DEFAULT_MAX_OPEN_FILES file handles simultaneously
    (e.g. when you are merging a lot of small images or PDFs). Equivalent of something like
    'default ulimit -n 1024' on macOS.

    NOTE: Does nothing on Windows (I think).
    NOTE: This mostly came from somewhere on stackoverflow but I lost the link.
    """
    try:
        import resource  # Windows doesn't have this package / doesn't need to bump up the ulimit (??)
    except ImportError:
        resource = None

    if resource is None:
        print_highlighted(f"No resource module; cannot set max open files on this platform...", style='yellow')
        return (None, None)
    elif num_filehandles <= DEFAULT_MAX_OPEN_FILES:
        # Then the OS max open files value is already sufficient.
        return (DEFAULT_MAX_OPEN_FILES, DEFAULT_MAX_OPEN_FILES)

    # %% (0) what is current ulimit -n setting?
    (soft, hard) = resource.getrlimit(resource.RLIMIT_NOFILE)
    num_filehandles = num_filehandles + OPEN_FILES_BUFFER

    # %% (1) increase limit (soft and even hard) if needed
    if soft < num_filehandles:
        soft = num_filehandles
        hard = max(soft, hard)
        print_highlighted(f"Increasing max open files soft & hard 'ulimit -n {soft} {hard}'...")

        try:
            resource.setrlimit(resource.RLIMIT_NOFILE, (soft, hard))
        except (ValueError, resource.error):
            try:
                hard = soft
                print_highlighted(f"Retrying setting max open files (soft, hard)=({soft}, {hard})", style='yellow')
                resource.setrlimit(resource.RLIMIT_NOFILE, (soft, hard))
            except Exception:
                print_highlighted('Failed to set max open files / ulimit, giving up!', style='error')
                soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)

    return (soft, hard)


def strip_bad_chars(text: str) -> str:
    """Remove chars that don't work well in filenames."""
    text = ' '.join(text.splitlines()).replace('\\s+', ' ')
    text = re.sub('’', "'", text).replace('|', 'I').replace(',', ',')
    return re.sub('[^-0-9a-zA-Z@.,?_:=#\'\\$" ()]+', '_', text).replace('  ', ' ')


def with_pdf_extension(file_path: Union[str, Path]) -> str:
    """Append `".pdf"` to `file_path` if it doesn't already end with `".pdf"`."""
    return str(file_path) + ('' if is_pdf(file_path) else PDF_EXT)



"""
Argument parsers for the command line tools other than `pdfalyze` that are included with The Pdfalyzer.

1. combine_pdfs
2. extract_pdf_pages
3. extract_pdf_text
"""
import logging
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path

from rich_argparse_plus import RichHelpFormatterPlus
from rich.prompt import Confirm
from rich.text import Text
from yaralyzer.helpers.file_helper import files_in_dir
from yaralyzer.util.logging import log

from pdfalyzer.util.argument_parser import ask_to_proceed, exit_with_error
from pdfalyzer.helpers.filesystem_helper import (do_all_files_exist, extract_page_number, file_exists, is_pdf,
     with_pdf_extension)
from pdfalyzer.helpers.rich_text_helper import print_highlighted
from pdfalyzer.util.page_range import PageRangeArgumentValidator

MAX_QUALITY = 10


##################
#  combine_pdfs  #
##################
combine_pdfs_parser = ArgumentParser(
    description="Combine multiple PDFs into one.",
    epilog="If all PDFs end in a number (e.g. 'xyz_1.pdf', 'xyz_2.pdf', etc. sort the files as if those were" +
           " page numbers prior to merging.",
    formatter_class=RichHelpFormatterPlus)

combine_pdfs_parser.add_argument('pdfs',
                                 help='two or more PDFs to combine',
                                 metavar='PDF_PATH',
                                 nargs='+')

combine_pdfs_parser.add_argument('-iq', '--image-quality',
                                 help='image quality for embedded images (can compress PDF at loss of quality)',
                                 choices=range(1, MAX_QUALITY + 1),
                                 default=MAX_QUALITY,
                                 type=int)

combine_pdfs_parser.add_argument('-o', '--output-file',
                                 help='path to write the combined PDFs to',
                                 required=True)


def parse_combine_pdfs_args() -> Namespace:
    """Parse command line args for combine_pdfs script."""
    args = combine_pdfs_parser.parse_args()
    args.output_file = with_pdf_extension(args.output_file)
    confirm_overwrite_txt = Text("Overwrite '").append(args.output_file, style='cyan').append("'?")
    args.number_of_pdfs = len(args.pdfs)

    if args.number_of_pdfs < 2:
        exit_with_error(f"Need at least 2 PDFs to merge.")
    elif not do_all_files_exist(args.pdfs):
        exit_with_error()
    elif file_exists(args.output_file) and not Confirm.ask(confirm_overwrite_txt):
        exit_with_error()

    if all(is_pdf(pdf) for pdf in args.pdfs):
        if all(extract_page_number(pdf) for pdf in args.pdfs):
            print_highlighted("PDFs appear to have page number suffixes so sorting numerically...")
            args.pdfs.sort(key=lambda pdf: extract_page_number(pdf))
        else:
            print_highlighted("PDFs don't seem to end in page numbers so using provided order...", style='yellow')
    else:
        print_highlighted("WARNING: At least one of the PDF args doesn't end in '.pdf'", style='bright_yellow')
        ask_to_proceed()

    print_highlighted(f"\nMerging {args.number_of_pdfs} individual PDFs into '{args.output_file}'...")
    return args


#####################
# extract_pdf_pages #
#####################
page_range_validator = PageRangeArgumentValidator()

extract_pdf_parser = ArgumentParser(
    formatter_class=RichHelpFormatterPlus,
    description="Extract pages from one PDF into a new PDF.",
)

extract_pdf_parser.add_argument('pdf_file', metavar='PDF_FILE', help='PDF to extract pages from')
extract_pdf_parser.add_argument('--debug', action='store_true', help='turn on debug level logging')

extract_pdf_parser.add_argument('--page-range', '-r',
                                type=page_range_validator,
                                help=page_range_validator.HELP_MSG,
                                required=True)

extract_pdf_parser.add_argument('--destination-dir', '-d',
                                help="directory to write the new PDF to",
                                default=Path.cwd())


def parse_pdf_page_extraction_args() -> Namespace:
    args = extract_pdf_parser.parse_args()

    if not is_pdf(args.pdf_file):
        log.error(f"'{args.pdf_file}' is not a PDF.")
        sys.exit(-1)
    elif not Path(args.destination_dir).exists():
        log.error(f"Destination dir '{args.destination_dir}' does not exist.")
        sys.exit(1)

    _set_log_level(args)
    return args


######################
#  extract_pdf_text  #
######################
extract_text_parser = ArgumentParser(
    formatter_class=RichHelpFormatterPlus,
    description="Extract the text from one or more files or directories.",
    epilog="If any of the FILE_OR_DIRs is a directory all PDF files in that directory will be extracted."
)

extract_text_parser.add_argument('file_or_dir', nargs='+', metavar='FILE_OR_DIR')
extract_text_parser.add_argument('--debug', action='store_true', help='turn on debug level logging')

extract_text_parser.add_argument('--page-range', '-r',
                                 type=page_range_validator,
                                 help=f"[PDFs only] {page_range_validator.HELP_MSG}")

extract_text_parser.add_argument('--print-as-parsed', '-p',
                                 action='store_true',
                                 help='print pages as they are parsed instead of waiting until parsing complete')


def parse_text_extraction_args() -> Namespace:
    args = extract_text_parser.parse_args()
    args.files_to_process = []

    for file_or_dir in args.file_or_dir:
        file_path = Path(file_or_dir)

        if not file_path.exists():
            log.error(f"'{file_path}' is not a valid file or directory.")
            sys.exit(-1)
        elif file_path.is_dir():
            args.files_to_process.extend(files_in_dir(file_path, 'pdf'))
        else:
            args.files_to_process.append(file_path)

    if args.page_range and (len(args.files_to_process) > 1 or not is_pdf(args.files_to_process[0])):
        log.error(f"--page-range can only be specified for a single PDF")
        sys.exit(-1)

    _set_log_level(args)
    return args


def _set_log_level(args: Namespace):
    if args.debug:
        log.setLevel(logging.DEBUG)



"""
Output formatting for font character mappings.
"""
from rich.columns import Columns
from rich.padding import Padding
from rich.text import Text
from yaralyzer.helpers.bytes_helper import print_bytes
from yaralyzer.output.rich_console import console
from yaralyzer.util.logging import log

# from pdfalyzer.font_info import FontInfo  # Causes circular import
from pdfalyzer.helpers.rich_text_helper import quoted_text
from pdfalyzer.helpers.string_helper import pp
from pdfalyzer.output.layout import print_headline_panel, subheading_width

CHARMAP_TITLE = 'Character Mapping (As Extracted By PyPDF)'
CHARMAP_INDENT = 8
CHARMAP_PADDING = (0, 2, 0, 10)
PREPARED_CHARMAP_PADDING = (0, 0, 0, CHARMAP_INDENT)
CHARMAP_TABLE_PADDING = (1, 0, 0, CHARMAP_INDENT + 2)
CHARMAP_PANEL_INTERNAL_INDENT = 2


def print_character_mapping(font: 'FontInfo') -> None:  # noqa: F821
    """Prints the character mapping extracted by PyPDF._charmap in tidy columns."""
    _print_charmap_header(f"{font} {CHARMAP_TITLE}", 'charmap.title')
    charmap_entries = [_format_charmap_entry(k, v) for k, v in font.character_mapping.items()]

    charmap_columns = Columns(
        charmap_entries,
        column_first=True,
        padding=CHARMAP_PADDING,
        equal=True,
        align='right'
    )

    console.print(Padding(charmap_columns, CHARMAP_TABLE_PADDING), width=subheading_width())
    console.line()


def print_prepared_charmap(font: 'FontInfo'):  # noqa: F821
    """Prints the prepared_charmap returned by PyPDF."""
    _print_charmap_header(f"{font} Adobe PostScript charmap prepared by PyPDF", 'charmap.prepared_title')
    console.line()
    print_bytes(font.prepared_char_map, style='charmap.prepared', indent=CHARMAP_INDENT + 2)
    console.line()


def _format_charmap_entry(k: str, v: str) -> Text:
    key = pp.pformat(k)

    for quote_char in ['"', "'"]:
        if len(key) > 1 and key.startswith(quote_char) and key.endswith(quote_char):
            key = key[1:-1]

    return quoted_text(key, 'charmap.byte') + Text(' => ') + quoted_text(str(v), 'charmap.char')


def _print_charmap_header(headline: str, style: str) -> None:
    right_padding = subheading_width() - CHARMAP_INDENT - CHARMAP_PANEL_INTERNAL_INDENT - len(headline) - 2
    print_headline_panel(headline, style, CHARMAP_INDENT, (0, right_padding, 0, 2))



"""
Parse command line arguments for `pdfalyze` and construct the `PdfalyzerConfig` object.
"""
import sys
from argparse import ArgumentParser, Namespace
from collections import namedtuple
from functools import partial, update_wrapper
from importlib.metadata import version
from os import getcwd, path
from typing import List, Optional

from rich_argparse_plus import RichHelpFormatterPlus
from rich.prompt import Confirm
from rich.text import Text
from yaralyzer.util.argument_parser import export, parser, parse_arguments as parse_yaralyzer_args, source
from yaralyzer.util.logging import log, log_and_print, log_argparse_result, log_current_config, log_invocation

from pdfalyzer.config import ALL_STREAMS, PDFALYZER, PdfalyzerConfig
from pdfalyzer.detection.constants.binary_regexes import QUOTE_PATTERNS
from pdfalyzer.helpers.rich_text_helper import print_highlighted

# NamedTuple to keep our argument selection orderly
OutputSection = namedtuple('OutputSection', ['argument', 'method'])

RichHelpFormatterPlus.choose_theme('prince')

DESCRIPTION = "Explore PDF's inner data structure with absurdly large and in depth visualizations. " + \
              "Track the control flow of her darker impulses, scan rivers of her binary data for signs " + \
              "of evil sorcery, and generally peer deep into the dark heart of the Portable Document Format. " + \
              "Just make sure you also forgive her - she knows not what she does."

EPILOG = "Values for various config options can be set permanently by a .pdfalyzer file in your home directory; " + \
         "see the documentation for details. " + \
         f"A registry of previous pdfalyzer invocations will be inscribed to a file if the " + \
         "{YaralyzerConfig.LOG_DIR_ENV_VAR} environment variable is configured."

# Analysis selection sections
DOCINFO = 'docinfo'
TREE = 'tree'
RICH = 'rich'
FONTS = 'fonts'
COUNTS = 'counts'
YARA = 'yara'
STREAMS = 'streams'
DEFAULT_SECTIONS = [DOCINFO, TREE, RICH, FONTS, COUNTS, YARA]
ALL_SECTIONS = DEFAULT_SECTIONS + [STREAMS]

# Add one more option to yaralyzer's export options
export.add_argument('-bin', '--extract-binary-streams',
                    action='store_const',
                    const='bin',
                    help='extract all binary streams in the PDF to separate files (requires pdf-parser.py)')

# Add one more option to the YARA rules section
source.add_argument('--no-default-yara-rules',
                    action='store_true',
                    help='if --yara is selected use only custom rules from --yara-file arg and not the default included YARA rules')


# Note that we extend the yaralyzer's parser and export
parser = ArgumentParser(
    formatter_class=RichHelpFormatterPlus,
    description=DESCRIPTION,
    epilog=EPILOG,
    parents=[parser],  # Extend yaralyzer args
    add_help=False)


# Output section selection
select = parser.add_argument_group(
    'ANALYSIS SELECTION',
    "Multiselect. Choosing nothing is choosing everything except --streams.")

select.add_argument('-d', '--docinfo', action='store_true',
                    help='show embedded document info (author, title, timestamps, etc.), streams overview, and MD5/SHA hashes')

select.add_argument('-t', '--tree', action='store_true',
                    help='show condensed tree visualization (one line per PDF object)')

select.add_argument('-r', '--rich', action='store_true',
                    help='show much larger / more detailed tree visualization (one row per PDF object property)')

select.add_argument('-f', '--fonts', action='store_true',
                    help="show info about fonts including character mappings for embedded font binaries")

select.add_argument('-y', '--yara', action='store_true',
                    help="scan the PDF with the included malicious PDF YARA rules and/or your custom YARA rules")

select.add_argument('-c', '--counts', action='store_true',
                    help='show counts of some of the properties of the objects in the PDF')

select.add_argument('-s', '--streams',
                    help="scan all the PDF's decoded/decrypted streams for sus content as well as any YARA rule matches. " +
                         "brute force is involved; output is verbose. a single OBJ_ID can be optionally provided to " +
                         "limit the output to a single internal object. try '-s -- [OTHERARGS]' if you run into an " +
                         "argument position related piccadilly.",
                    nargs='?',
                    const=ALL_STREAMS,
                    metavar='ID',
                    type=int)

select.add_argument('--extract-quoted',
                    help="extract and force decode all bytes found between this kind of quotation marks " +
                         "(requires --streams. can be specified more than once)",
                    choices=list(QUOTE_PATTERNS.keys()),
                    dest='extract_quoteds',
                    action='append')

select.add_argument('--suppress-boms', action='store_true',
                    help="don't scan streams for byte order marks (suppresses some of the --streams output)")

select.add_argument('--preview-stream-length',
                    help='number of bytes at the beginning and end of stream data to show as a preview',
                    metavar='BYTES',
                    type=int)

# Make sure the selection section is at the top
parser._action_groups = parser._action_groups[:2] + [parser._action_groups[-1]] + parser._action_groups[2:-1]


################################
# Main argument parsing begins #
################################
def parse_arguments():
    """Parse command line args. Most args can also be communicated to the app by setting env vars."""
    if '--version' in sys.argv:
        print(f"pdfalyzer {version(PDFALYZER)}")
        sys.exit()

    args = parser.parse_args()
    args = parse_yaralyzer_args(args)
    log_invocation()

    if not args.streams:
        if args.extract_quoteds:
            exit_with_error("--extract-quoted does nothing if --streams is not selected")
        if args.suppress_boms:
            log.warning("--suppress-boms has nothing to suppress if --streams is not selected")

    if args.no_default_yara_rules and not args.yara_rules_files:
        exit_with_error("--no-default-yara-rules requires at least one --yara-file argument")

    # File export options
    if args.export_svg or args.export_txt or args.export_html or args.extract_binary_streams:
        args.output_dir = args.output_dir or getcwd()
        file_prefix = (args.file_prefix + '__') if args.file_prefix else ''
        args.file_suffix = ('_' + args.file_suffix) if args.file_suffix else ''
        args.output_basename = f"{file_prefix}{path.basename(args.file_to_scan_path)}"
    elif args.output_dir:
        log.warning('--output-dir provided but no export option was chosen')

    args.extract_quoteds = args.extract_quoteds or []
    PdfalyzerConfig._args = args
    log_argparse_result(args, 'parsed')
    log_current_config()
    return args


def output_sections(args: Namespace, pdfalyzer: 'Pdfalyzer') -> List[OutputSection]:  # noqa: F821
    """
    Determine which of the tree visualizations, font scans, etc should be run.
    If nothing is specified output ALL sections other than --streams which is v. slow/verbose.

    Args:
        args: parsed command line arguments
        pdfalyzer: the `pdfalyzer` instance whose methods will be called to produce output
    Returns:
        List[OutputSection]: List of `OutputSection` namedtuples with 'argument' and 'method' fields
    """
    # Create a partial for print_font_info() because it's the only one that can take an argument
    # partials have no __name__ so update_wrapper() propagates the 'print_font_info' as this partial's name
    stream_id = None if args.streams == ALL_STREAMS else args.streams
    stream_scan = partial(pdfalyzer.print_streams_analysis, idnum=stream_id)
    update_wrapper(stream_scan, pdfalyzer.print_streams_analysis)

    # 1st element string matches the argument in 'select' group
    # 2nd is fxn to call if selected.
    # Top to bottom is the default order of output.
    possible_output_sections = [
        OutputSection(DOCINFO, pdfalyzer.print_document_info),
        OutputSection(TREE, pdfalyzer.print_tree),
        OutputSection(RICH, pdfalyzer.print_rich_table_tree),
        OutputSection(FONTS, pdfalyzer.print_font_info),
        OutputSection(COUNTS, pdfalyzer.print_summary),
        OutputSection(YARA, pdfalyzer.print_yara_results),
        OutputSection(STREAMS, stream_scan),
    ]

    output_sections = [section for section in possible_output_sections if vars(args)[section.argument]]

    if len(output_sections) == 0:
        log_and_print("No output section specified so outputting all sections except --streams...")
        return [section for section in possible_output_sections if section.argument != STREAMS]
    else:
        return output_sections


def all_sections_chosen(args):
    """Returns True if all flags are set or no flags are set."""
    return len([s for s in ALL_SECTIONS if vars(args)[s]]) == len(ALL_SECTIONS)


#############
#  Helpers  #
#############

def ask_to_proceed() -> None:
    """Exit if user doesn't confirm they want to proceed."""
    if not Confirm.ask(Text("Proceed anyway?")):
        exit_with_error()


def exit_with_error(error_message: Optional[str] = None) -> None:
    """Print 'error_message' and exit with status code 1."""
    if error_message:
        print_highlighted(Text('').append('ERROR', style='bold red').append(f': {error_message}'))

    print_highlighted('Exiting...', style='dim red')
    sys.exit(1)



"""
Verify that the PDF tree is complete/contains all the nodes in the PDF file.
"""
from dataclasses import dataclass

from pypdf.errors import PdfReadError
from pypdf.generic import IndirectObject, NameObject, NumberObject
from rich.markup import escape
from yaralyzer.output.rich_console import console
from yaralyzer.util.logging import log

from pdfalyzer.util.adobe_strings import *


@dataclass
class PdfTreeVerifier:
    """Class to verify that the PDF tree is complete/contains all the nodes in the PDF file."""
    pdfalyzer: 'Pdfalyzer'

    def verify_all_nodes_encountered_are_in_tree(self) -> None:
        """Make sure every node we can see is reachable from the root of the tree"""
        missing_nodes = [
            node for idnum, node in self.pdfalyzer.nodes_encountered.items()
            if self.pdfalyzer.find_node_by_idnum(idnum) is None
        ]

        if len(missing_nodes) > 0:
            msg = f"Nodes were traversed but never placed: {escape(str(missing_nodes))}\n" + \
                   "For link nodes like /First, /Next, /Prev, and /Last this might be no big deal - depends " + \
                   "on the PDF. But for other node typtes this could indicate missing data in the tree."
            console.print(msg)
            log.warning(msg)

    def verify_unencountered_are_untraversable(self) -> None:
        """Make sure any PDF object IDs we can't find in tree are /ObjStm or /Xref nodes."""
        if self.pdfalyzer.pdf_size is None:
            log.warning(f"{SIZE} not found in PDF trailer; cannot verify all nodes are in tree")
            return
        if self.pdfalyzer.max_generation > 0:
            log.warning(f"Methodd doesn't check revisions but this doc is generation {self.pdfalyzer.max_generation}")

        # We expect to see all ordinals up to the number of nodes /Trailer claims exist as obj. IDs.
        missing_node_ids = [
            i for i in range(1, self.pdfalyzer.pdf_size)
            if self.pdfalyzer.find_node_by_idnum(i) is None
        ]

        for idnum in missing_node_ids:
            ref = IndirectObject(idnum, self.pdfalyzer.max_generation, self.pdfalyzer.pdf_reader)

            try:
                obj = ref.get_object()
            except PdfReadError as e:
                if 'Invalid Elementary Object' in str(e):
                    log.warning(f"Couldn't verify elementary obj with id {idnum} is properly in tree")
                    continue
                log.error(str(e))
                console.print_exception()
                obj = None
                raise e

            if obj is None:
                log.error(f"Cannot find ref {ref} in PDF!")
                continue
            elif isinstance(obj, (NumberObject, NameObject)):
                log.info(f"Obj {idnum} is a {type(obj)} w/value {obj}; if relationshipd by /Length etc. this is a nonissue but maybe worth doublechecking")  # noqa: E501
                continue
            elif not isinstance(obj, dict):
                log.error(f"Obj {idnum} ({obj}) of type {type(obj)} isn't dict, cannot determine if it should be in tree")  # noqa: E501
                continue
            elif TYPE not in obj:
                msg = f"Obj {idnum} has no {TYPE} and is not in tree. Either a loose node w/no data or an error in pdfalyzer."  # noqa: E501
                msg += f"\nHere's the contents for you to assess:\n{obj}"
                log.warning(msg)
                continue

            obj_type = obj[TYPE]

            if obj_type == OBJECT_STREAM:
                log.debug(f"Object with id {idnum} not found in tree because it's an {OBJECT_STREAM}")
            elif obj[TYPE] == XREF:
                placeable = XREF_STREAM in self.pdfalyzer.pdf_reader.trailer

                for k, v in self.pdfalyzer.pdf_reader.trailer.items():
                    xref_val_for_key = obj.get(k)

                    if k in [XREF_STREAM, PREV]:
                        continue
                    elif k == SIZE:
                        if xref_val_for_key is None or v != (xref_val_for_key + 1):
                            log.info(f"{XREF} has {SIZE} of {xref_val_for_key}, trailer has {SIZE} of {v}")
                            placeable = False

                        continue
                    elif k not in obj or v != obj.get(k):
                        log.info(f"Trailer has {k} -> {v} but {XREF} obj has {obj.get(k)} at that key")
                        placeable = False

                if placeable:
                    self.pdfalyzer.pdf_tree.add_child(self.pdfalyzer._build_or_find_node(ref, XREF_STREAM))
            else:
                log.warning(f"{XREF} Obj {idnum} not found in tree!")



from typing import Optional

from yaralyzer.output.rich_console import console

from pdfalyzer.helpers.rich_text_helper import print_warning


def ocr_text(image: "Image.Image", image_name: str) -> Optional[str]:  # noqa F821
    """Use pytesseract to OCR the text in the image and return it as a string."""
    import pytesseract
    from PIL import Image
    text = None

    try:
        text = pytesseract.image_to_string(image)
    except pytesseract.pytesseract.TesseractError:
        console.print_exception()
        print_warning(f"Tesseract OCR failure '{image_name}'! No OCR text extracted...")
    except OSError as e:
        if 'truncated' in str(e):
            print_warning(f"Truncated image file '{image_name}'!")
        else:
            console.print_exception()
            console.print(f"Error while extracting '{image_name}'!", style='bright_red')
            raise e
    except Exception as e:
        console.print_exception()
        console.print(f"Error while extracting '{image_name}'!", style='bright_red')
        raise e

    return None if text is None else text.strip()



"""
Configuration of what to scan for in binary data. Regexes here will be matched against binary streams
and then force decoded.
"""
from pdfalyzer.util.adobe_strings import DANGEROUS_PDF_KEYS

DANGEROUS_JAVASCRIPT_INSTRUCTIONS = ['eval']
DANGEROUS_PDF_KEYS_TO_HUNT_WITH_SLASH = ['/URI']
DANGEROUS_PDF_KEYS_TO_HUNT_ONLY_IN_FONTS = ['/AA', '/F']

HEX_ENCODED_PDF_KEYS_TO_HUNT = [
    '/#55#52#49',  # URI
]

# Potentially dangerous PDF instructions: Remove the leading '/' and convert to bytes except /F ("URL")
DANGEROUS_STRINGS = [instruction[1:] for instruction in DANGEROUS_PDF_KEYS]
DANGEROUS_STRINGS.extend(DANGEROUS_PDF_KEYS_TO_HUNT_WITH_SLASH)
DANGEROUS_STRINGS.extend(DANGEROUS_JAVASCRIPT_INSTRUCTIONS)

# Quote capture regexes
DOUBLE_QUOTE = 'double_quote'
SINGLE_QUOTE = 'single_quote'

BACKSLASH = 'backslash'
BACKTICK = 'backtick'
BRACKET = 'bracket'
CURLY_BRACKET = 'curly_bracket'
DOUBLE_LESS_THAN = 'double_lessthan'
ESCAPED_SINGLE = f"escaped_{SINGLE_QUOTE}"
ESCAPED_DOUBLE = f"escaped_{DOUBLE_QUOTE}"
FRONTSLASH = 'frontslash'
GUILLEMET = 'guillemet'
LESS_THAN = 'lessthan'
PARENTHESES = 'parentheses'


QUOTE_PATTERNS = {
    BACKTICK: '`.+`',
    BRACKET: '\\[.+\\]',         # { 91 [-] 93 }
    CURLY_BRACKET: '{.+}',       # { 123 [-] 125 }
    DOUBLE_LESS_THAN: '<<.+>>',  # Hex { 60 60 [-] 62 62 }
    ESCAPED_SINGLE: "\\'.+\\'",
    ESCAPED_DOUBLE: '\\".+\\"',
    FRONTSLASH: '/.+/',          # { 47 [-] 47 }
    GUILLEMET: 'AB [-] BB',      # Guillemet quotes are not ANSI so require byte pattern
    LESS_THAN: '<.+>',           # Hex { 60 [-] 62 }
    PARENTHESES: '\\(.+\\)',     # Hex { 28 [-] 29 }
}



HIGH_VALUE_KEYWORDS = [
    'this', # *
    'require ', # *
    'const ', # *
    'eval',  # *
]


JAVASCRIPT_RESERVED_KEYWORDS = HIGH_VALUE_KEYWORDS + [
    'abstract',
    'arguments',
    'await',
    'boolean',
    'break',
    'byte',
    'case',
    'catch',
    'char',
    'class',
    'continue',
    'debugger',
    'default',
    'delete',
    'do',
    'double',
    'else',
    'enum',
    'export',
    'extends',
    'false',
    'final',
    'finally',
    'float',
    'for',
    'function',
    'goto',
    'if',
    'implements',
    'import', # *
    'in',
    'instanceof',
    'int',
    'interface',
    'let',
    'long',
    'native',
    'new',
    'null',
    'package',
    'private',
    'protected',
    'public',
    'return',
    'short',
    'static',
    'super',
    'switch',
    'synchronized',
    'throw',
    'throws',
    'transient',
    'true',
    'try',
    'typeof',
    'var',
    'void',
    'volatile',
    'while',
    'with',
    'yield'
]



"""
Configurations and methods to help with consistent styling of PDF trees, nodes, etc.
"""
import re
from collections import namedtuple
from numbers import Number
from typing import Any

from pypdf.generic import (ArrayObject, ByteStringObject, EncodedStreamObject, IndirectObject,
     StreamObject, TextStringObject)
from yaralyzer.output.rich_console import YARALYZER_THEME_DICT

from pdfalyzer.output.styles.rich_theme import PDF_ARRAY
from pdfalyzer.util import adobe_strings

ClassStyle = namedtuple('ClassStyle', ['klass', 'style'])

DEFAULT_LABEL_STYLE = 'yellow'
FONT_OBJ_BLUE = 'deep_sky_blue4 bold'
PDF_NON_TREE_REF = 'color(243)'

# Subclasses of the key type will be styled with the value string
NODE_TYPE_STYLES = [
    ClassStyle(Number, 'cyan bold'),
    ClassStyle(IndirectObject, 'color(225)'),
    ClassStyle(ByteStringObject, 'bytes'),
    ClassStyle(EncodedStreamObject, YARALYZER_THEME_DICT['bytes']),
    ClassStyle(StreamObject, YARALYZER_THEME_DICT['bytes.title']),
    ClassStyle(TextStringObject, YARALYZER_THEME_DICT['grey.light']),
    ClassStyle(ArrayObject, PDF_ARRAY),
    ClassStyle(dict, 'color(64)'),
    ClassStyle(list, 'color(143)'),
    ClassStyle(str, 'bright_white bold'),
    # Default
    ClassStyle(object, 'bright_yellow'),
]

LABEL_STYLES = [
    [re.compile('JavaScript|JS|OpenAction', re.I | re.M), 'blink bold red'],
    [re.compile(f'^{adobe_strings.FONT_DESCRIPTOR}'),     'cornflower_blue'],
    [re.compile(f'^{adobe_strings.FONT_FILE}'),           'steel_blue1'],
    [re.compile(f'^{adobe_strings.FONT}'),                FONT_OBJ_BLUE],
    [re.compile(f'^{adobe_strings.TO_UNICODE}'),          'grey30'],
    [re.compile(f'^{adobe_strings.ENCODING}'),            YARALYZER_THEME_DICT['encoding.header']],
    [re.compile(f'^{adobe_strings.WIDTHS}'),              'color(67)'],
    [re.compile(f'^{adobe_strings.W}'),                   'color(67)'],
    [re.compile(f'^{adobe_strings.RESOURCES}'),           'magenta'],
    [re.compile('/(Trailer|Root|Info|Outlines)'),         'bright_green'],
    [re.compile('/Catalog'),                              'color(47)'],
    [re.compile('/(Metadata|ViewerPreferences)'),         'color(35)'],
    [re.compile('^/Contents'),                            'medium_purple1'],
    [re.compile('^/Action'),                              'dark_red'],
    [re.compile('^/Annots'),                              'deep_sky_blue4'],
    [re.compile('^/Annot'),                               'color(24)'],
    [re.compile('^/Pages'),                               'dark_orange3'],
    [re.compile('^/Page'),                                'light_salmon3'],
    [re.compile('^/ColorSpace'),                          'medium_orchid1'],
    [re.compile('^/(URI|Names)'),                         'white'],
    [re.compile(f'^{adobe_strings.XOBJECT}'),             'grey37'],
    [re.compile(f'^{adobe_strings.UNLABELED}'),           'grey35 reverse'],
    [re.compile(f'^{adobe_strings.XREF}'),                'color(148)'],
]

# Add common color for all NON_TREE_REFERENCES
LABEL_STYLES += [
    [re.compile(f'^{key}'), PDF_NON_TREE_REF]
    for key in adobe_strings.NON_TREE_REFERENCES
]


def get_class_style(obj: Any) -> str:
    """Style for various types of data (e.g. DictionaryObject)"""
    return next((cs.style for cs in NODE_TYPE_STYLES if isinstance(obj, cs.klass)), '')


def get_class_style_dim(obj: Any) -> str:
    """Dim version of get_class_style() for non primitives, white for primitives"""
    if isinstance(obj, str):
        return 'color(244)'
    elif isinstance(obj, Number):
        return 'cyan dim'
    else:
        return f"{get_class_style(obj)} dim"


def get_class_style_italic(obj: Any) -> str:
    return f"{get_class_style(obj)} italic"


def get_label_style(label: str) -> str:
    """Lookup a style based on the node's label string (either its type or first address)."""
    return next((ls[1] for ls in LABEL_STYLES if ls[0].search(label)), DEFAULT_LABEL_STYLE)



"""
Count the Javascript (at least the 3+ letter words, record big matches.
"""
import re

from pdfalyzer.detection.constants.javascript_reserved_keywords import JAVASCRIPT_RESERVED_KEYWORDS
from pdfalyzer.helpers.string_helper import count_regex_matches_in_text

JS_KEYWORDS_3_OR_MORE_LETTERS = [kw for kw in JAVASCRIPT_RESERVED_KEYWORDS if len(kw) > 2]
JS_KEYWORD_REGEX = re.compile('|'.join(JS_KEYWORDS_3_OR_MORE_LETTERS))


class JavascriptHunter:
    @classmethod
    def count_js_keywords_in_text(cls, text: str) -> int:
        return count_regex_matches_in_text(JS_KEYWORD_REGEX, text)

    @classmethod
    def js_keyword_matches(cls, text: str) -> [str]:
        return JS_KEYWORD_REGEX.findall(text)



"""
PdfalyzerConfig object holds the unification of configuration options parsed from the command line
as well as those set by environment variables and/or a .pdfalyzer file.
"""
import importlib.resources
from argparse import Namespace
from os import environ, pardir, path

from yaralyzer.config import YaralyzerConfig, is_env_var_set_and_not_false, is_invoked_by_pytest

PDFALYZE = 'pdfalyze'
PDFALYZER = f"{PDFALYZE}r"
ALL_STREAMS = -1
PYTEST_FLAG = 'INVOKED_BY_PYTEST'
PROJECT_ROOT = path.join(str(importlib.resources.files(PDFALYZER)), pardir)

# 3rd part pdf-parser.py
PDF_PARSER_EXECUTABLE_ENV_VAR = 'PDFALYZER_PDF_PARSER_PY_PATH'
DEFAULT_PDF_PARSER_EXECUTABLE = path.join(PROJECT_ROOT, 'tools', 'pdf-parser.py')


class PdfalyzerConfig:
    _args: Namespace = Namespace()

    # Path to Didier Stevens's pdf-parser.py
    if is_env_var_set_and_not_false(PDF_PARSER_EXECUTABLE_ENV_VAR):
        PDF_PARSER_EXECUTABLE = path.join(environ[PDF_PARSER_EXECUTABLE_ENV_VAR], 'pdf-parser.py')
    elif is_invoked_by_pytest():
        PDF_PARSER_EXECUTABLE = DEFAULT_PDF_PARSER_EXECUTABLE
    else:
        if path.exists(DEFAULT_PDF_PARSER_EXECUTABLE):
            PDF_PARSER_EXECUTABLE = DEFAULT_PDF_PARSER_EXECUTABLE
        else:
            PDF_PARSER_EXECUTABLE = None

    @classmethod
    def get_output_basepath(cls, export_method: str) -> str:
        """Build the path to an output file - everything but the extension"""
        export_type = export_method.__name__.removeprefix('print_')
        output_basename = f"{cls._args.output_basename}.{export_type}"

        if export_type == 'streams_analysis':
            if cls._args.streams != ALL_STREAMS:
                output_basename += f"_streamid{cls._args.streams}"

            output_basename += f"_maxdecode{YaralyzerConfig.args.max_decode_length}"

            if cls._args.extract_quoteds:
                output_basename += f"_extractquoteds-{','.join(cls._args.extract_quoteds)}"

        return path.join(
            cls._args.output_dir,
            output_basename + cls._args.file_suffix + f"___pdfalyzed_{cls._args.invoked_at_str}"
        )



"""
`PdfTreeNode` decorates a `PdfObject` with tree structure information.
"""
from typing import Callable, List, Optional

from anytree import NodeMixin, SymlinkNode
from pypdf.errors import PdfReadError
from pypdf.generic import IndirectObject, PdfObject, StreamObject
from rich.markup import escape
from rich.text import Text
from yaralyzer.output.rich_console import console
from yaralyzer.util.logging import log

from pdfalyzer.decorators.pdf_object_properties import PdfObjectProperties
from pdfalyzer.helpers.string_helper import is_prefixed_by_any
from pdfalyzer.pdf_object_relationship import PdfObjectRelationship
from pdfalyzer.util.adobe_strings import *
from pdfalyzer.util.exceptions import PdfWalkError

DEFAULT_MAX_ADDRESS_LENGTH = 90
DECODE_FAILURE_LEN = -1


class PdfTreeNode(NodeMixin, PdfObjectProperties):
    """
    PDF node decorator - wraps actual PDF objects to make them `anytree` nodes.
    Also adds decorators/generators for Rich text representation.

    Child/parent relationships should be set using the `add_child()` and `set_parent()`
    methods and not set directly.

    TODO: this could be done better with anytree hooks.
    """

    def __init__(self, obj: PdfObject, address: str, idnum: int):
        """
        Args:
            obj (PdfObject): The underlying PDF object
            address (str): The first address that points from some node to this one
            idnum (int): ID used in the reference
        """
        PdfObjectProperties.__init__(self, obj, address, idnum)
        self.non_tree_relationships: List[PdfObjectRelationship] = []

        if isinstance(obj, StreamObject):
            try:
                self.stream_data = self.obj.get_data()
                self.stream_length = len(self.stream_data)
            except (NotImplementedError, PdfReadError) as e:
                msg = f"PyPDF failed to decode stream in {self}: {e}.\n" + \
                       "Trees will be unaffected but scans/extractions will not be able to check this stream."
                console.print_exception()
                log.warning(msg)
                console.print(msg, style='error')
                self.stream_data = msg.encode()
                self.stream_length = DECODE_FAILURE_LEN
        else:
            self.stream_data = None
            self.stream_length = 0

    @classmethod
    def from_reference(cls, ref: IndirectObject, address: str) -> 'PdfTreeNode':
        """Alternate constructor to Build a `PdfTreeNode` from an `IndirectObject`."""
        try:
            return cls(ref.get_object(), address, ref.idnum)
        except PdfReadError as e:
            console.print_exception()
            msg = f"Failed to build node properly because of above exception ({e}). " + \
                   "Tree integrity not guaranteed."
            log.warning(msg)
            return cls(ref, address, ref.idnum)

    def set_parent(self, parent: 'PdfTreeNode') -> None:
        """Set the parent of this node."""
        if self.parent is not None and self.parent != parent:
            raise PdfWalkError(f"Cannot set {parent} as parent of {self}, parent is already {self.parent}")

        self.parent = parent
        self.remove_non_tree_relationship(parent)
        self.known_to_parent_as = self.address_of_this_node_in_other(parent) or self.first_address
        log.info(f"  Added {parent} as parent of {self}")

    def add_child(self, child: 'PdfTreeNode') -> None:
        """Add a child to this node."""
        if next((c for c in self.children if c.idnum == child.idnum), None) is not None:
            log.debug(f"{child} is already child of {self}")
        else:
            child.set_parent(self)

    def add_non_tree_relationship(self, relationship: PdfObjectRelationship) -> None:
        """Add a relationship that points at this node's PDF object. TODO: doesn't include parent/child."""
        if relationship in self.non_tree_relationships:
            return

        self.non_tree_relationships.append(relationship)
        log.info(f'Added other relationship: {relationship} {self}')

    def remove_non_tree_relationship(self, from_node: 'PdfTreeNode') -> None:
        """Remove all non_tree_relationships from from_node to this node."""
        relationships_to_remove = [r for r in self.non_tree_relationships if r.from_node == from_node]

        if len(relationships_to_remove) == 0:
            return
        elif len(relationships_to_remove) > 1 and \
                not all(r.reference_key in [FIRST, LAST] for r in relationships_to_remove):
            log.warning(f"> 1 relationships to remove from {from_node} to {self}: {relationships_to_remove}")

        for relationship in relationships_to_remove:
            log.debug(f"Removing relationship {relationship} from {self}")
            self.non_tree_relationships.remove(relationship)

    def nodes_with_here_references(self) -> List['PdfTreeNode']:
        """Return a list of nodes that contain this node's PDF object as an IndirectObject reference."""
        return [r.from_node for r in self.non_tree_relationships if r.from_node]

    def non_tree_relationship_count(self) -> int:
        """Number of non parent/child relationships containing this node."""
        return len(self.non_tree_relationships)

    def unique_addresses(self) -> List[str]:
        """All the addresses in other nodes that refer to this object."""
        addresses = set([r.address for r in self.non_tree_relationships])

        if self.known_to_parent_as is not None:
            addresses.add(self.known_to_parent_as)

        return list(addresses)

    def references_to_other_nodes(self) -> List[PdfObjectRelationship]:
        """Returns all nodes referenced from node.obj (see PdfObjectRelationship definition)."""
        return PdfObjectRelationship.build_node_references(from_node=self)

    def contains_stream(self) -> bool:
        """Returns True for ContentStream, DecodedStream, and EncodedStream objects."""
        return isinstance(self.obj, StreamObject)

    def tree_address(self, max_length: Optional[int] = DEFAULT_MAX_ADDRESS_LENGTH) -> str:
        """Creates a string like '/Catalog/Pages/Resources[2]/Font' truncated to max_length (if given)."""
        if self.label == TRAILER:
            return '/'
        elif self.parent is None:
            raise PdfWalkError(f"{self} does not have a parent; cannot get accurate node address.")
        elif self.parent.label == TRAILER:
            return self.known_to_parent_as

        address = self.parent.tree_address() + self.known_to_parent_as

        if max_length is None or max_length > len(address):
            return address

        return '...' + address[-max_length:][3:]

    def address_of_this_node_in_other(self, from_node: 'PdfTreeNode') -> Optional[str]:
        """Find the local address used in 'from_node' to refer to this node."""
        refs_to_this_node = [
            ref for ref in from_node.references_to_other_nodes()
            if ref.to_obj.idnum == self.idnum
        ]

        if len(refs_to_this_node) == 1:
            return refs_to_this_node[0].address
        elif len(refs_to_this_node) == 0:
            # TODO: Hack city. /XRef streams are basically trailer nodes without any direct reference
            if self.parent and self.parent.label == TRAILER and self.type == XREF and XREF_STREAM in self.parent.obj:
                return XREF_STREAM
            elif self.label not in NON_STANDARD_ADDRESS_NODES:
                log.info(f"Could not find expected reference from {from_node} to {self}")
            else:
                return None
        else:
            address = refs_to_this_node[0].address

            # If other node's label doesn't start with a NON_STANDARD_ADDRESS string
            # AND any of the relationships pointing at this node use something other than a
            #     NON_STANDARD_ADDRESS_NODES string to refer here,
            # then print a warning about multiple refs.
            if not (is_prefixed_by_any(from_node.label, NON_STANDARD_ADDRESS_NODES)
                    or
                    all(ref.address in NON_STANDARD_ADDRESS_NODES for ref in refs_to_this_node)):
                refs_to_this_node_str = "\n   ".join([f"{i + 1}. {r}" for i, r in enumerate(refs_to_this_node)])
                msg = f"Multiple refs from {from_node} to {self}:\n   {refs_to_this_node_str}"
                log.warning(msg + f"\nCommon address of refs: {address}")

            return address

    def tree_relationships(self) -> List['PdfTreeNode']:
        """Returns parents and children."""
        return list(self.children) + ([self.parent] if self.parent is not None else [])

    def symlink_non_tree_relationships(self):
        """Create SymlinkNodes for this node's non parent/child (non-tree) relationships."""
        log.info(f"Symlinking {self}'s {self.non_tree_relationship_count()} other relationships...")

        for relationship in self.non_tree_relationships:
            if relationship.from_node in self.tree_relationships():
                log.warning(f"  {relationship} is still 'non-tree' but is a parent or child of {self}")
            else:
                log.debug(f"   SymLinking {relationship} to {self}")
                SymlinkNode(self, parent=relationship.from_node)

    def descendants_count(self) -> int:
        """Count nodes in the tree that are children/grandchildren/great grandchildren/etc of this one."""
        return len(self.children) + sum([child.descendants_count() for child in self.children])

    def unique_labels_of_referring_nodes(self) -> List[str]:
        """Unique label strings of nodes referring here outside the parent/child hierarchy."""
        return list(set([r.from_node.label for r in self.non_tree_relationships]))

    def print_non_tree_relationships(self) -> None:
        """console.print this node's non tree relationships (represented by SymlinkNodes in the tree)."""
        self._write_non_tree_relationships(console.print)

    def log_non_tree_relationships(self) -> None:
        """log this node's non tree relationships (represented by SymlinkNodes in the tree)."""
        self._write_non_tree_relationships(log.warning)

    def _write_non_tree_relationships(self, write_method: Callable) -> None:
        """Use write_method() to write self.non_tree_relationships."""
        write_method(f"{escape(str(self))} parent from candidates:")

        for i, r in enumerate(self.non_tree_relationships):
            write_method(f"  {i + 1}. {escape(str(r))}, Descendant Count: {r.from_node.descendants_count()}")

    def _colored_address(self, max_length: Optional[int] = None) -> Text:
        """Rich text version of tree_address()."""
        text = Text('@', style='bright_white')
        return text.append(self.tree_address(max_length), style='address')

    def __rich__(self) -> Text:
        return PdfObjectProperties.__rich__(self)[:-1] + self._colored_address() + Text('>')

    def __str__(self) -> str:
        return PdfObjectProperties.__rich__(self).plain

    def __repr__(self) -> str:
        return self.__str__()



"""
Rich colors: https://rich.readthedocs.io/en/stable/appendix/colors.html
TODO: interesting colors # row_styles[0] = 'reverse bold on color(144)' <-
"""
from rich.theme import Theme
from yaralyzer.output.rich_console import GREY_ADDRESS, YARALYZER_THEME_DICT, console

# Colors / PDF object styles
DANGER_HEADER = 'color(88) on white'  # Red
PDF_ARRAY = 'color(120)'
PDF_NON_TREE_REF = 'color(243)'

PDFALYZER_THEME_DICT = YARALYZER_THEME_DICT.copy()

PDFALYZER_THEME_DICT.update({
    'address': GREY_ADDRESS,
    'BOM': 'bright_green',
    # PDF objects
    'pdf.array': PDF_ARRAY,
    'pdf.non_tree_ref': PDF_NON_TREE_REF,
    # fonts
    'font.property': 'color(135)',
    'font.title': 'reverse dark_blue on color(253)',
    # charmap
    'charmap.title': 'color(18) reverse on white dim',
    'charmap.prepared_title': 'color(23) reverse on white dim',
    'charmap.prepared': 'color(106) dim',
    'charmap.byte': 'color(58)',
    'charmap.char': 'color(120) bold',
    # design elements
    'subtable': 'color(8) on color(232)',
    'header.minor': 'color(249) bold',
    'header.danger': DANGER_HEADER,
    'header.danger_reverse': f'{DANGER_HEADER} reverse',
    # neutral log events
    'event.attn': 'bold bright_cyan',
    'event.lowpriority': 'bright_black',
    # good log events
    'event.good': 'green4',
    'event.better': 'turquoise4',
    'event.reallygood': 'dark_cyan',
    'event.reallygreat': 'spring_green1',
    'event.great': 'sea_green2',
    'event.evenbetter': 'chartreuse1',
    'event.best': 'green1',
    'event.siren': 'blink bright_white on red3',
    # warn log events
    'warn': 'bright_yellow',
    'warn.mild': 'yellow2',
    'warn.milder': 'dark_orange3',
    'warn.harsh': 'reverse bright_yellow',
    # error log events
    'fail': 'bold reverse red',
    'mild_error': 'red',  # TODO: unused?
    'mild_warning': 'color(228) dim',
    'red_alert': 'blink bold red reverse on white',
})

# Override whatever theme The Yaralyzer has configured.
console.push_theme(Theme(PDFALYZER_THEME_DICT))
